# coding: utf-8

"""
    Lightly API

    Lightly.ai enables you to do self-supervised learning in an easy and intuitive way. The lightly.ai OpenAPI spec defines how one can interact with our REST API to unleash the full potential of lightly.ai  # noqa: E501

    OpenAPI spec version: 1.0.0
    Contact: support@lightly.ai
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


import pprint
import re  # noqa: F401

import six

from lightly.openapi_generated.swagger_client.configuration import Configuration


class DockerTaskDescription(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'embeddings_filename': 'str',
        'embeddings_hash': 'str',
        'method': 'SamplingMethod',
        'existing_selection_column_name': 'str',
        'active_learning_scores_column_name': 'str',
        'masked_out_column_name': 'str',
        'sampling_config': 'SamplingConfig',
        'n_data': 'float'
    }

    attribute_map = {
        'embeddings_filename': 'embeddingsFilename',
        'embeddings_hash': 'embeddingsHash',
        'method': 'method',
        'existing_selection_column_name': 'existingSelectionColumnName',
        'active_learning_scores_column_name': 'activeLearningScoresColumnName',
        'masked_out_column_name': 'maskedOutColumnName',
        'sampling_config': 'samplingConfig',
        'n_data': 'nData'
    }

    def __init__(self, embeddings_filename=None, embeddings_hash=None, method=None, existing_selection_column_name=None, active_learning_scores_column_name=None, masked_out_column_name=None, sampling_config=None, n_data=None, _configuration=None):  # noqa: E501
        """DockerTaskDescription - a model defined in Swagger"""  # noqa: E501
        if _configuration is None:
            _configuration = Configuration()
        self._configuration = _configuration

        self._embeddings_filename = None
        self._embeddings_hash = None
        self._method = None
        self._existing_selection_column_name = None
        self._active_learning_scores_column_name = None
        self._masked_out_column_name = None
        self._sampling_config = None
        self._n_data = None
        self.discriminator = None

        self.embeddings_filename = embeddings_filename
        self.embeddings_hash = embeddings_hash
        self.method = method
        self.existing_selection_column_name = existing_selection_column_name
        self.active_learning_scores_column_name = active_learning_scores_column_name
        self.masked_out_column_name = masked_out_column_name
        self.sampling_config = sampling_config
        self.n_data = n_data

    @property
    def embeddings_filename(self):
        """Gets the embeddings_filename of this DockerTaskDescription.  # noqa: E501


        :return: The embeddings_filename of this DockerTaskDescription.  # noqa: E501
        :rtype: str
        """
        return self._embeddings_filename

    @embeddings_filename.setter
    def embeddings_filename(self, embeddings_filename):
        """Sets the embeddings_filename of this DockerTaskDescription.


        :param embeddings_filename: The embeddings_filename of this DockerTaskDescription.  # noqa: E501
        :type: str
        """
        if self._configuration.client_side_validation and embeddings_filename is None:
            raise ValueError("Invalid value for `embeddings_filename`, must not be `None`")  # noqa: E501

        self._embeddings_filename = embeddings_filename

    @property
    def embeddings_hash(self):
        """Gets the embeddings_hash of this DockerTaskDescription.  # noqa: E501


        :return: The embeddings_hash of this DockerTaskDescription.  # noqa: E501
        :rtype: str
        """
        return self._embeddings_hash

    @embeddings_hash.setter
    def embeddings_hash(self, embeddings_hash):
        """Sets the embeddings_hash of this DockerTaskDescription.


        :param embeddings_hash: The embeddings_hash of this DockerTaskDescription.  # noqa: E501
        :type: str
        """
        if self._configuration.client_side_validation and embeddings_hash is None:
            raise ValueError("Invalid value for `embeddings_hash`, must not be `None`")  # noqa: E501

        self._embeddings_hash = embeddings_hash

    @property
    def method(self):
        """Gets the method of this DockerTaskDescription.  # noqa: E501


        :return: The method of this DockerTaskDescription.  # noqa: E501
        :rtype: SamplingMethod
        """
        return self._method

    @method.setter
    def method(self, method):
        """Sets the method of this DockerTaskDescription.


        :param method: The method of this DockerTaskDescription.  # noqa: E501
        :type: SamplingMethod
        """
        if self._configuration.client_side_validation and method is None:
            raise ValueError("Invalid value for `method`, must not be `None`")  # noqa: E501

        self._method = method

    @property
    def existing_selection_column_name(self):
        """Gets the existing_selection_column_name of this DockerTaskDescription.  # noqa: E501


        :return: The existing_selection_column_name of this DockerTaskDescription.  # noqa: E501
        :rtype: str
        """
        return self._existing_selection_column_name

    @existing_selection_column_name.setter
    def existing_selection_column_name(self, existing_selection_column_name):
        """Sets the existing_selection_column_name of this DockerTaskDescription.


        :param existing_selection_column_name: The existing_selection_column_name of this DockerTaskDescription.  # noqa: E501
        :type: str
        """
        if self._configuration.client_side_validation and existing_selection_column_name is None:
            raise ValueError("Invalid value for `existing_selection_column_name`, must not be `None`")  # noqa: E501

        self._existing_selection_column_name = existing_selection_column_name

    @property
    def active_learning_scores_column_name(self):
        """Gets the active_learning_scores_column_name of this DockerTaskDescription.  # noqa: E501


        :return: The active_learning_scores_column_name of this DockerTaskDescription.  # noqa: E501
        :rtype: str
        """
        return self._active_learning_scores_column_name

    @active_learning_scores_column_name.setter
    def active_learning_scores_column_name(self, active_learning_scores_column_name):
        """Sets the active_learning_scores_column_name of this DockerTaskDescription.


        :param active_learning_scores_column_name: The active_learning_scores_column_name of this DockerTaskDescription.  # noqa: E501
        :type: str
        """
        if self._configuration.client_side_validation and active_learning_scores_column_name is None:
            raise ValueError("Invalid value for `active_learning_scores_column_name`, must not be `None`")  # noqa: E501

        self._active_learning_scores_column_name = active_learning_scores_column_name

    @property
    def masked_out_column_name(self):
        """Gets the masked_out_column_name of this DockerTaskDescription.  # noqa: E501


        :return: The masked_out_column_name of this DockerTaskDescription.  # noqa: E501
        :rtype: str
        """
        return self._masked_out_column_name

    @masked_out_column_name.setter
    def masked_out_column_name(self, masked_out_column_name):
        """Sets the masked_out_column_name of this DockerTaskDescription.


        :param masked_out_column_name: The masked_out_column_name of this DockerTaskDescription.  # noqa: E501
        :type: str
        """
        if self._configuration.client_side_validation and masked_out_column_name is None:
            raise ValueError("Invalid value for `masked_out_column_name`, must not be `None`")  # noqa: E501

        self._masked_out_column_name = masked_out_column_name

    @property
    def sampling_config(self):
        """Gets the sampling_config of this DockerTaskDescription.  # noqa: E501


        :return: The sampling_config of this DockerTaskDescription.  # noqa: E501
        :rtype: SamplingConfig
        """
        return self._sampling_config

    @sampling_config.setter
    def sampling_config(self, sampling_config):
        """Sets the sampling_config of this DockerTaskDescription.


        :param sampling_config: The sampling_config of this DockerTaskDescription.  # noqa: E501
        :type: SamplingConfig
        """
        if self._configuration.client_side_validation and sampling_config is None:
            raise ValueError("Invalid value for `sampling_config`, must not be `None`")  # noqa: E501

        self._sampling_config = sampling_config

    @property
    def n_data(self):
        """Gets the n_data of this DockerTaskDescription.  # noqa: E501

        the number of samples in the current embeddings file  # noqa: E501

        :return: The n_data of this DockerTaskDescription.  # noqa: E501
        :rtype: float
        """
        return self._n_data

    @n_data.setter
    def n_data(self, n_data):
        """Sets the n_data of this DockerTaskDescription.

        the number of samples in the current embeddings file  # noqa: E501

        :param n_data: The n_data of this DockerTaskDescription.  # noqa: E501
        :type: float
        """
        if self._configuration.client_side_validation and n_data is None:
            raise ValueError("Invalid value for `n_data`, must not be `None`")  # noqa: E501

        self._n_data = n_data

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(DockerTaskDescription, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, DockerTaskDescription):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, DockerTaskDescription):
            return True

        return self.to_dict() != other.to_dict()



from concurrent.futures import ThreadPoolExecutor
from typing import List, Mapping, Optional, Sequence, Tuple

import tqdm

from lightly.api.prediction_singletons import PredictionSingletonRepr
from lightly.openapi_generated.swagger_client import PredictionTaskSchema


class _PredictionsMixin:
    def create_or_update_prediction_task_schema(
        self,
        schema: PredictionTaskSchema,
        prediction_version_id: int = -1,
    ) -> None:
        """Creates or updates the prediction task schema

        Args:
            schema:
                The prediction task schema.
            prediction_version_id:
                A numerical id (e.g timestamp) to distinguish different predictions of different model versions.
                Use the same id if you don't require versioning or if you wish to overwrite the previous schema.

        Example:
          >>> import time
          >>> from lightly.api import ApiWorkflowClient
          >>> from lightly.openapi_generated.swagger_client import (
          >>>     PredictionTaskSchema,
          >>>     TaskType,
          >>>     PredictionTaskSchemaCategory,
          >>> )
          >>>
          >>> client = ApiWorkflowClient(
          >>>     token="MY_LIGHTLY_TOKEN", dataset_id="MY_DATASET_ID"
          >>> )
          >>>
          >>> schema = PredictionTaskSchema(
          >>>     name="my-object-detection",
          >>>     type=TaskType.OBJECT_DETECTION,
          >>>     categories=[
          >>>         PredictionTaskSchemaCategory(id=0, name="dog"),
          >>>         PredictionTaskSchemaCategory(id=1, name="cat"),
          >>>     ],
          >>> )
          >>> client.create_or_update_prediction_task_schema(schema=schema)


        """
        self._predictions_api.create_or_update_prediction_task_schema_by_dataset_id(
            body=schema,
            dataset_id=self.dataset_id,
            prediction_uuid_timestamp=prediction_version_id,
        )

    def create_or_update_predictions(
        self,
        sample_id_to_prediction_singletons: Mapping[str, Sequence[PredictionSingletonRepr]],
        prediction_version_id: int = -1,
        progress_bar: Optional[tqdm.tqdm] = None,
        max_workers: int = 8,
    ) -> None:
        """Creates or updates the predictions for specific samples

        Args:
            sample_id_to_prediction_singletons
                A mapping from the sample_id of the sample to its corresponding prediction singletons.
                The singletons can be from different tasks and different types.

            prediction_version_id:
                 A numerical id (e.g timestamp) to distinguish different predictions of different model versions.
                 Use the same id if you don't require versioning or if you wish to overwrite the previous schema.
                 This id must match the id of a prediction task schema.

            progress_bar:
                Tqdm progress bar to show how many prediction files have already been uploaded.

            max_workers:
                Maximum number of workers uploading predictions in parallel.

        Example:
          >>> import time
          >>> from tqdm import tqdm
          >>> from lightly.api import ApiWorkflowClient
          >>> from lightly.openapi_generated.swagger_client import (
          >>>     PredictionTaskSchema,
          >>>     TaskType,
          >>>     PredictionTaskSchemaCategory,
          >>> )
          >>> from lightly.api.prediction_singletons import PredictionSingletonClassificationRepr
          >>>
          >>> client = ApiWorkflowClient(
          >>>     token="MY_LIGHTLY_TOKEN", dataset_id="MY_DATASET_ID"
          >>> )
          >>>
          >>> samples = client._samples_api.get_samples_partial_by_dataset_id(dataset_id=client.dataset_id, mode=SamplePartialMode.FILENAMES)
          >>> sample_id_to_prediction_singletons_dummy = {
          >>>     sample.id: [PredictionSingletonClassificationRepr(taskName="my-task", categoryId=i%4, score=0.9, probabilities=[0.1, 0.2, 0.3, 0.4])]
          >>>     for i, sample in enumerate(samples)
          >>> }
          >>> client.create_or_update_predictions(
          >>>     sample_id_to_prediction_singletons=sample_id_to_prediction_singletons_dummy,
          >>>     progress_bar=tqdm(desc="Uploading predictions", total=len(samples), unit=" predictions")
          >>> )


        """

        # handle the case where len(sample_id_to_prediction_singletons) < max_workers
        max_workers = min(len(sample_id_to_prediction_singletons), max_workers)
        max_workers = max(max_workers, 1)

        def upload_prediction(
            sample_id_prediction_singletons_tuple: Tuple[
                str, Sequence[PredictionSingletonRepr]
            ]
        ) -> None:
            (sample_id, prediction_singletons) = sample_id_prediction_singletons_tuple
            self.create_or_update_prediction(
                sample_id=sample_id,
                prediction_singletons=prediction_singletons,
                prediction_version_id=prediction_version_id,
            )

        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            for _ in executor.map(
                upload_prediction, sample_id_to_prediction_singletons.items()
            ):
                if progress_bar is not None:
                    progress_bar.update(1)

    def create_or_update_prediction(
        self,
        sample_id: str,
        prediction_singletons: Sequence[PredictionSingletonRepr],
        prediction_version_id: int = -1,
    ) -> None:
        """Creates or updates the predictions for one specific sample

        Args:
            sample_id
                The id of the sample

            prediction_version_id:
                And id to distinguish different predictions for the same sample.

            prediction_singletons:
                The predictions to upload for that sample
        """
        prediction_singletons_for_sending = [
            singleton.to_dict() for singleton in prediction_singletons
        ]
        self._predictions_api.create_or_update_prediction_by_sample_id(
            body=prediction_singletons_for_sending,
            dataset_id=self.dataset_id,
            sample_id=sample_id,
            prediction_uuid_timestamp=prediction_version_id,
        )



import math

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.distributed as dist


def prototype_probabilities(
    queries: torch.Tensor,
    prototypes: torch.Tensor,
    temperature: float,
) -> torch.Tensor:
    """Returns probability for each query to belong to each prototype.

    Args:
        queries:
            Tensor with shape (batch_size, dim)
        prototypes:
            Tensor with shape (num_prototypes, dim)
        temperature:
            Inverse scaling factor for the similarity.

    Returns:
        Probability tensor with shape (batch_size, num_prototypes) which sums to 1 along
        the num_prototypes dimension.

    """
    return F.softmax(torch.matmul(queries, prototypes.T) / temperature, dim=1)

def sharpen(probabilities: torch.Tensor, temperature: float) -> torch.Tensor:
    """Sharpens the probabilities with the given temperature.

    Args:
        probabilities:
            Tensor with shape (batch_size, dim)
        temperature:
            Temperature in (0, 1]. Lower temperature results in stronger sharpening (
            output probabilities are less uniform).
    Returns:
        Probabilities tensor with shape (batch_size, dim).

    """
    probabilities = probabilities ** (1.0 / temperature)
    probabilities /= torch.sum(probabilities, dim=1, keepdim=True)
    return probabilities

@torch.no_grad()
def sinkhorn(
    probabilities: torch.Tensor,
    iterations: int = 3,
    gather_distributed: bool = False,
) -> torch.Tensor:
    """Runs sinkhorn normalization on the probabilities as described in [0].

    Code inspired by [1].

    - [0]: Masked Siamese Networks, 2022, https://arxiv.org/abs/2204.07141
    - [1]: https://github.com/facebookresearch/msn

    Args:
        probabilities:
            Probabilities tensor with shape (batch_size, num_prototypes).
        iterations:
            Number of iterations of the sinkhorn algorithms. Set to 0 to disable.
        gather_distributed:
            If True then features from all gpus are gathered during normalization.
    Returns:
        A normalized probabilities tensor.

    """
    if iterations <= 0:
        return probabilities

    world_size = 1
    if gather_distributed and dist.is_initialized():
        world_size = dist.get_world_size()

    num_targets, num_prototypes = probabilities.shape
    probabilities = probabilities.T
    sum_probabilities = torch.sum(probabilities)
    if world_size > 1:
        dist.all_reduce(sum_probabilities)
    probabilities = probabilities / sum_probabilities

    for _ in range(iterations):
        # normalize rows
        row_sum = torch.sum(probabilities, dim=1, keepdim=True)
        if world_size > 1:
            dist.all_reduce(row_sum)
        probabilities /= row_sum
        probabilities /= num_prototypes

        # normalize columns
        probabilities /= torch.sum(probabilities, dim=0, keepdim=True)
        probabilities /= num_targets

    probabilities *= num_targets
    return probabilities.T


class MSNLoss(nn.Module):
    """Implementation of the loss function from MSN [0].

    Code inspired by [1].

    - [0]: Masked Siamese Networks, 2022, https://arxiv.org/abs/2204.07141
    - [1]: https://github.com/facebookresearch/msn

    Attributes:
        temperature:
            Similarities between anchors and targets are scaled by the inverse of
            the temperature. Must be in (0, 1].
        sinkhorn_iterations:
            Number of sinkhorn normalization iterations on the targets.
        me_max_weight:
            Weight factor lambda by which the mean entropy maximization regularization
            loss is scaled. Set to 0 to disable the reguliarization.

     Examples:

        >>> # initialize loss function
        >>> loss_fn = MSNLoss()
        >>>
        >>> # generate anchors and targets of images
        >>> anchors = transforms(images)
        >>> targets = transforms(images)
        >>>
        >>> # feed through MSN model
        >>> anchors_out = model(anchors)
        >>> targets_out = model.target(targets)
        >>>
        >>> # calculate loss
        >>> loss = loss_fn(anchors_out, targets_out, prototypes=model.prototypes)

    """
    def __init__(
        self,
        temperature: float = 0.1,
        sinkhorn_iterations: int = 3,
        me_max_weight: float = 1.0,
        gather_distributed: bool = False,
    ):
        super().__init__()
        self.temperature = temperature
        self.sinkhorn_iterations = sinkhorn_iterations
        self.me_max_weight = me_max_weight
        self.gather_distributed = gather_distributed

    def forward(
        self,
        anchors: torch.Tensor,
        targets: torch.Tensor,
        prototypes: torch.Tensor,
        target_sharpen_temperature: float = 0.25,
    ) -> torch.Tensor:
        """Computes the MSN loss for a set of anchors, targets and prototypes.

        Args:
            anchors:
                Tensor with shape (batch_size * anchor_views, dim).
            targets:
                Tensor with shape (batch_size, dim).
            prototypes:
                Tensor with shape (num_prototypes, dim).
            target_sharpen_temperature:
                Temperature used to sharpen the target probabilities.

        Returns:
            Mean loss over all anchors.

        """
        num_views = anchors.shape[0] // targets.shape[0]
        anchors = F.normalize(anchors, dim=1)
        targets = F.normalize(targets, dim=1)
        prototypes = F.normalize(prototypes, dim=1)

        # anchor predictions
        anchor_probs = prototype_probabilities(anchors, prototypes, temperature=self.temperature)

        # target predictions
        with torch.no_grad():
            target_probs = prototype_probabilities(targets, prototypes, temperature=self.temperature)
            target_probs = sharpen(target_probs, temperature=target_sharpen_temperature)
            if self.sinkhorn_iterations > 0:
                target_probs = sinkhorn(
                    probabilities=target_probs,
                    iterations=self.sinkhorn_iterations,
                    gather_distributed=self.gather_distributed
                )
            target_probs = target_probs.repeat((num_views, 1))

        # cross entropy loss
        loss = torch.mean(torch.sum(torch.log(anchor_probs**(-target_probs)), dim=1))

        # mean entropy maximization regularization
        if self.me_max_weight > 0:
            mean_anchor_probs = torch.mean(anchor_probs, dim=0)
            me_max_loss = torch.sum(torch.log(mean_anchor_probs**(-mean_anchor_probs)))
            me_max_loss += math.log(float(len(mean_anchor_probs)))
            loss -= self.me_max_weight * me_max_loss

        return loss



# coding: utf-8

"""
    Lightly API

    Lightly.ai enables you to do self-supervised learning in an easy and intuitive way. The lightly.ai OpenAPI spec defines how one can interact with our REST API to unleash the full potential of lightly.ai  # noqa: E501

    OpenAPI spec version: 1.0.0
    Contact: support@lightly.ai
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


import pprint
import re  # noqa: F401

import six

from lightly.openapi_generated.swagger_client.configuration import Configuration


class SampleMetaData(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'custom': 'dict(str, object)',
        'dynamic': 'dict(str, object)',
        'sharpness': 'float',
        'size_in_bytes': 'int',
        'snr': 'float',
        'mean': 'list[float]',
        'shape': 'list[int]',
        'std': 'list[float]',
        'sum_of_squares': 'list[float]',
        'sum_of_values': 'list[float]'
    }

    attribute_map = {
        'custom': 'custom',
        'dynamic': 'dynamic',
        'sharpness': 'sharpness',
        'size_in_bytes': 'sizeInBytes',
        'snr': 'snr',
        'mean': 'mean',
        'shape': 'shape',
        'std': 'std',
        'sum_of_squares': 'sumOfSquares',
        'sum_of_values': 'sumOfValues'
    }

    def __init__(self, custom=None, dynamic=None, sharpness=None, size_in_bytes=None, snr=None, mean=None, shape=None, std=None, sum_of_squares=None, sum_of_values=None, _configuration=None):  # noqa: E501
        """SampleMetaData - a model defined in Swagger"""  # noqa: E501
        if _configuration is None:
            _configuration = Configuration()
        self._configuration = _configuration

        self._custom = None
        self._dynamic = None
        self._sharpness = None
        self._size_in_bytes = None
        self._snr = None
        self._mean = None
        self._shape = None
        self._std = None
        self._sum_of_squares = None
        self._sum_of_values = None
        self.discriminator = None

        if custom is not None:
            self.custom = custom
        if dynamic is not None:
            self.dynamic = dynamic
        if sharpness is not None:
            self.sharpness = sharpness
        if size_in_bytes is not None:
            self.size_in_bytes = size_in_bytes
        if snr is not None:
            self.snr = snr
        if mean is not None:
            self.mean = mean
        if shape is not None:
            self.shape = shape
        if std is not None:
            self.std = std
        if sum_of_squares is not None:
            self.sum_of_squares = sum_of_squares
        if sum_of_values is not None:
            self.sum_of_values = sum_of_values

    @property
    def custom(self):
        """Gets the custom of this SampleMetaData.  # noqa: E501


        :return: The custom of this SampleMetaData.  # noqa: E501
        :rtype: dict(str, object)
        """
        return self._custom

    @custom.setter
    def custom(self, custom):
        """Sets the custom of this SampleMetaData.


        :param custom: The custom of this SampleMetaData.  # noqa: E501
        :type: dict(str, object)
        """

        self._custom = custom

    @property
    def dynamic(self):
        """Gets the dynamic of this SampleMetaData.  # noqa: E501


        :return: The dynamic of this SampleMetaData.  # noqa: E501
        :rtype: dict(str, object)
        """
        return self._dynamic

    @dynamic.setter
    def dynamic(self, dynamic):
        """Sets the dynamic of this SampleMetaData.


        :param dynamic: The dynamic of this SampleMetaData.  # noqa: E501
        :type: dict(str, object)
        """

        self._dynamic = dynamic

    @property
    def sharpness(self):
        """Gets the sharpness of this SampleMetaData.  # noqa: E501


        :return: The sharpness of this SampleMetaData.  # noqa: E501
        :rtype: float
        """
        return self._sharpness

    @sharpness.setter
    def sharpness(self, sharpness):
        """Sets the sharpness of this SampleMetaData.


        :param sharpness: The sharpness of this SampleMetaData.  # noqa: E501
        :type: float
        """

        self._sharpness = sharpness

    @property
    def size_in_bytes(self):
        """Gets the size_in_bytes of this SampleMetaData.  # noqa: E501


        :return: The size_in_bytes of this SampleMetaData.  # noqa: E501
        :rtype: int
        """
        return self._size_in_bytes

    @size_in_bytes.setter
    def size_in_bytes(self, size_in_bytes):
        """Sets the size_in_bytes of this SampleMetaData.


        :param size_in_bytes: The size_in_bytes of this SampleMetaData.  # noqa: E501
        :type: int
        """

        self._size_in_bytes = size_in_bytes

    @property
    def snr(self):
        """Gets the snr of this SampleMetaData.  # noqa: E501


        :return: The snr of this SampleMetaData.  # noqa: E501
        :rtype: float
        """
        return self._snr

    @snr.setter
    def snr(self, snr):
        """Sets the snr of this SampleMetaData.


        :param snr: The snr of this SampleMetaData.  # noqa: E501
        :type: float
        """

        self._snr = snr

    @property
    def mean(self):
        """Gets the mean of this SampleMetaData.  # noqa: E501


        :return: The mean of this SampleMetaData.  # noqa: E501
        :rtype: list[float]
        """
        return self._mean

    @mean.setter
    def mean(self, mean):
        """Sets the mean of this SampleMetaData.


        :param mean: The mean of this SampleMetaData.  # noqa: E501
        :type: list[float]
        """

        self._mean = mean

    @property
    def shape(self):
        """Gets the shape of this SampleMetaData.  # noqa: E501


        :return: The shape of this SampleMetaData.  # noqa: E501
        :rtype: list[int]
        """
        return self._shape

    @shape.setter
    def shape(self, shape):
        """Sets the shape of this SampleMetaData.


        :param shape: The shape of this SampleMetaData.  # noqa: E501
        :type: list[int]
        """

        self._shape = shape

    @property
    def std(self):
        """Gets the std of this SampleMetaData.  # noqa: E501


        :return: The std of this SampleMetaData.  # noqa: E501
        :rtype: list[float]
        """
        return self._std

    @std.setter
    def std(self, std):
        """Sets the std of this SampleMetaData.


        :param std: The std of this SampleMetaData.  # noqa: E501
        :type: list[float]
        """

        self._std = std

    @property
    def sum_of_squares(self):
        """Gets the sum_of_squares of this SampleMetaData.  # noqa: E501


        :return: The sum_of_squares of this SampleMetaData.  # noqa: E501
        :rtype: list[float]
        """
        return self._sum_of_squares

    @sum_of_squares.setter
    def sum_of_squares(self, sum_of_squares):
        """Sets the sum_of_squares of this SampleMetaData.


        :param sum_of_squares: The sum_of_squares of this SampleMetaData.  # noqa: E501
        :type: list[float]
        """

        self._sum_of_squares = sum_of_squares

    @property
    def sum_of_values(self):
        """Gets the sum_of_values of this SampleMetaData.  # noqa: E501


        :return: The sum_of_values of this SampleMetaData.  # noqa: E501
        :rtype: list[float]
        """
        return self._sum_of_values

    @sum_of_values.setter
    def sum_of_values(self, sum_of_values):
        """Sets the sum_of_values of this SampleMetaData.


        :param sum_of_values: The sum_of_values of this SampleMetaData.  # noqa: E501
        :type: list[float]
        """

        self._sum_of_values = sum_of_values

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(SampleMetaData, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, SampleMetaData):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, SampleMetaData):
            return True

        return self.to_dict() != other.to_dict()



""" SplitBatchNorm Implementation """

# Copyright (c) 2020. Lightly AG and its affiliates.
# All Rights Reserved

import torch
import torch.nn as nn


class SplitBatchNorm(nn.BatchNorm2d):
    """Simulates multi-gpu behaviour of BatchNorm in one gpu by splitting.

    Implementation was adapted from:
    https://github.com/davidcpage/cifar10-fast/blob/master/torch_backend.py

    Attributes:
        num_features:
            Number of input features.
        num_splits:
            Number of splits.

    """
    def __init__(self, num_features, num_splits, **kw):
        super().__init__(num_features, **kw)
        self.num_splits = num_splits
        self.register_buffer(
            'running_mean', torch.zeros(num_features*self.num_splits)
        )
        self.register_buffer(
            'running_var', torch.ones(num_features*self.num_splits)
        )

    def train(self, mode=True):
        # lazily collate stats when we are going to use them
        if (self.training is True) and (mode is False):
            self.running_mean = \
                torch.mean(
                    self.running_mean.view(self.num_splits, self.num_features),
                    dim=0
                ).repeat(self.num_splits)
            self.running_var = \
                torch.mean(
                    self.running_var.view(self.num_splits, self.num_features),
                    dim=0
                ).repeat(self.num_splits)

        return super().train(mode)

    def forward(self, input):
        """Computes the SplitBatchNorm on the input.

        """
        # get input shape
        N, C, H, W = input.shape

        # during training, use different stats for each split and otherwise
        # use the stats from the first split
        if self.training or not self.track_running_stats:
            result = nn.functional.batch_norm(
                input.view(-1, C*self.num_splits, H, W),
                self.running_mean, self.running_var, 
                self.weight.repeat(self.num_splits),
                self.bias.repeat(self.num_splits),
                True,
                self.momentum,
                self.eps
            ).view(N, C, H, W)
        else:
            result = nn.functional.batch_norm(
                input,
                self.running_mean[:self.num_features],
                self.running_var[:self.num_features], 
                self.weight,
                self.bias,
                False,
                self.momentum, 
                self.eps
            )
        
        return result


def get_norm_layer(num_features: int, num_splits: int, **kw):
    """Utility to switch between BatchNorm2d and SplitBatchNorm.

    """
    if num_splits > 0:
        return SplitBatchNorm(num_features, num_splits)
    else:
        return nn.BatchNorm2d(num_features)



# coding: utf-8

"""
    Lightly API

    Lightly.ai enables you to do self-supervised learning in an easy and intuitive way. The lightly.ai OpenAPI spec defines how one can interact with our REST API to unleash the full potential of lightly.ai  # noqa: E501

    OpenAPI spec version: 1.0.0
    Contact: support@lightly.ai
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


import pprint
import re  # noqa: F401

import six

from lightly.openapi_generated.swagger_client.configuration import Configuration


class Embedding2dCreateRequest(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'name': 'str',
        'dimensionality_reduction_method': 'DimensionalityReductionMethod',
        'coordinates_dimension1': 'Embedding2dCoordinates',
        'coordinates_dimension2': 'Embedding2dCoordinates'
    }

    attribute_map = {
        'name': 'name',
        'dimensionality_reduction_method': 'dimensionalityReductionMethod',
        'coordinates_dimension1': 'coordinatesDimension1',
        'coordinates_dimension2': 'coordinatesDimension2'
    }

    def __init__(self, name=None, dimensionality_reduction_method=None, coordinates_dimension1=None, coordinates_dimension2=None, _configuration=None):  # noqa: E501
        """Embedding2dCreateRequest - a model defined in Swagger"""  # noqa: E501
        if _configuration is None:
            _configuration = Configuration()
        self._configuration = _configuration

        self._name = None
        self._dimensionality_reduction_method = None
        self._coordinates_dimension1 = None
        self._coordinates_dimension2 = None
        self.discriminator = None

        self.name = name
        self.dimensionality_reduction_method = dimensionality_reduction_method
        self.coordinates_dimension1 = coordinates_dimension1
        self.coordinates_dimension2 = coordinates_dimension2

    @property
    def name(self):
        """Gets the name of this Embedding2dCreateRequest.  # noqa: E501

        Name of the 2d embedding (default is embedding name + __2d)  # noqa: E501

        :return: The name of this Embedding2dCreateRequest.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this Embedding2dCreateRequest.

        Name of the 2d embedding (default is embedding name + __2d)  # noqa: E501

        :param name: The name of this Embedding2dCreateRequest.  # noqa: E501
        :type: str
        """
        if self._configuration.client_side_validation and name is None:
            raise ValueError("Invalid value for `name`, must not be `None`")  # noqa: E501

        self._name = name

    @property
    def dimensionality_reduction_method(self):
        """Gets the dimensionality_reduction_method of this Embedding2dCreateRequest.  # noqa: E501


        :return: The dimensionality_reduction_method of this Embedding2dCreateRequest.  # noqa: E501
        :rtype: DimensionalityReductionMethod
        """
        return self._dimensionality_reduction_method

    @dimensionality_reduction_method.setter
    def dimensionality_reduction_method(self, dimensionality_reduction_method):
        """Sets the dimensionality_reduction_method of this Embedding2dCreateRequest.


        :param dimensionality_reduction_method: The dimensionality_reduction_method of this Embedding2dCreateRequest.  # noqa: E501
        :type: DimensionalityReductionMethod
        """
        if self._configuration.client_side_validation and dimensionality_reduction_method is None:
            raise ValueError("Invalid value for `dimensionality_reduction_method`, must not be `None`")  # noqa: E501

        self._dimensionality_reduction_method = dimensionality_reduction_method

    @property
    def coordinates_dimension1(self):
        """Gets the coordinates_dimension1 of this Embedding2dCreateRequest.  # noqa: E501


        :return: The coordinates_dimension1 of this Embedding2dCreateRequest.  # noqa: E501
        :rtype: Embedding2dCoordinates
        """
        return self._coordinates_dimension1

    @coordinates_dimension1.setter
    def coordinates_dimension1(self, coordinates_dimension1):
        """Sets the coordinates_dimension1 of this Embedding2dCreateRequest.


        :param coordinates_dimension1: The coordinates_dimension1 of this Embedding2dCreateRequest.  # noqa: E501
        :type: Embedding2dCoordinates
        """
        if self._configuration.client_side_validation and coordinates_dimension1 is None:
            raise ValueError("Invalid value for `coordinates_dimension1`, must not be `None`")  # noqa: E501

        self._coordinates_dimension1 = coordinates_dimension1

    @property
    def coordinates_dimension2(self):
        """Gets the coordinates_dimension2 of this Embedding2dCreateRequest.  # noqa: E501


        :return: The coordinates_dimension2 of this Embedding2dCreateRequest.  # noqa: E501
        :rtype: Embedding2dCoordinates
        """
        return self._coordinates_dimension2

    @coordinates_dimension2.setter
    def coordinates_dimension2(self, coordinates_dimension2):
        """Sets the coordinates_dimension2 of this Embedding2dCreateRequest.


        :param coordinates_dimension2: The coordinates_dimension2 of this Embedding2dCreateRequest.  # noqa: E501
        :type: Embedding2dCoordinates
        """
        if self._configuration.client_side_validation and coordinates_dimension2 is None:
            raise ValueError("Invalid value for `coordinates_dimension2`, must not be `None`")  # noqa: E501

        self._coordinates_dimension2 = coordinates_dimension2

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(Embedding2dCreateRequest, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, Embedding2dCreateRequest):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, Embedding2dCreateRequest):
            return True

        return self.to_dict() != other.to_dict()



"""
FIXME: hypersphere is perhaps bad naming as I am not sure it is the essence;
 alignment-and-uniformity loss perhaps? Does not sound as nice.
"""

import torch
import torch.nn.functional as F


class HypersphereLoss(torch.nn.Module):
    """
    Implementation of the loss described in 'Understanding Contrastive Representation Learning through
    Alignment and Uniformity on the Hypersphere.' [0]
    
    [0] Tongzhou Wang. et.al, 2020, ... https://arxiv.org/abs/2005.10242

    Note:
        In order for this loss to function as advertized, an l1-normalization to the hypersphere is required.
        This loss function applies this l1-normalization internally in the loss-layer.
        However, it is recommended that the same normalization is also applied in your architecture,
        considering that this l1-loss is also intended to be applied during inference.
        Perhaps there may be merit in leaving it out of the inferrence pathway, but this use has not been tested.

        Moreover it is recommended that the layers preceeding this loss function are either a linear layer without activation,
        a batch-normalization layer, or both. The directly upstream architecture can have a large influence
        on the ability of this loss to achieve its stated aim of promoting uniformity on the hypersphere;
        and if by contrast the last layer going into the embedding is a RELU or similar nonlinearity,
        we may see that we will never get very close to achieving the goal of uniformity on the hypersphere,
        but will confine ourselves to the subspace of positive activations.
        Similar architectural considerations are relevant to most contrastive loss functions,
        but we call it out here explicitly.

    Examples:
        >>> # initialize loss function
        >>> loss_fn = HypersphereLoss()
        >>>
        >>> # generate two random transforms of images
        >>> t0 = transforms(images)
        >>> t1 = transforms(images)
        >>>
        >>> # feed through SimSiam model
        >>> out0, out1 = model(t0, t1)
        >>>
        >>> # calculate loss
        >>> loss = loss_fn(out0, out1)

    """

    def __init__(self, t=1., lam=1., alpha=2.):
        """Parameters as described in [0]

        Args:
            t : float
                Temperature parameter;
                proportional to the inverse variance of the Gaussians used to measure uniformity
            lam : float:
                Weight balancing the alignment and uniformity loss terms
            alpha : float
                Power applied to the alignment term of the loss. At its default value of 2,
                distances between positive samples are penalized in an l-2 sense.

        """
        super(HypersphereLoss, self).__init__()
        self.t = t
        self.lam = lam
        self.alpha = alpha

    def forward(self, z_a: torch.Tensor, z_b: torch.Tensor) -> torch.Tensor:
        """

        Args:
            x (torch.Tensor, [b, d], float)
            y (torch.Tensor, [b, d], float)

        Returns:
            Loss (torch.Tensor, [], float)

        """
        x = F.normalize(z_a)
        y = F.normalize(z_b)

        def lalign(x, y):
            return (x - y).norm(dim=1).pow(self.alpha).mean()
        def lunif(x):
            sq_pdist = torch.pdist(x, p=2).pow(2)
            return sq_pdist.mul(-self.t).exp().mean().log()
        return lalign(x, y) + self.lam * (lunif(x) + lunif(y)) / 2



""" Solarization """

# Copyright (c) 2021. Lightly AG and its affiliates.
# All Rights Reserved

import numpy as np
from PIL import ImageOps


class RandomSolarization(object):
    """Implementation of random image Solarization.

    Utilizes the integrated image operation `solarize` from Pillow. Solarization
    inverts all pixel values above a threshold (default: 128).

    Attributes:
        probability:
            Probability to apply the transformation
        threshold:
            Threshold for solarization.
    """

    def __init__(self, prob: float = 0.5,
                 threshold: int = 128):
        self.prob = prob
        self.threshold = threshold

    def __call__(self, sample):
        """Solarizes the given input image

        Args:
            sample:
                PIL image to which solarize will be applied.

        Returns:
            Solarized image or original image.

        """
        prob = np.random.random_sample()
        if prob < self.prob:
            # return solarized image
            return ImageOps.solarize(sample, threshold=self.threshold)
        # return original image
        return sample



import warnings
from typing import *

import numpy as np

from lightly.active_learning.scorers.scorer import Scorer


def _entropy(probs: np.ndarray, axis: int = 1) -> np.ndarray:
    """Computes the entropy of a probability matrix over one array

    Args:
        probs:
            A probability matrix of shape (N, M)
        axis:
            The axis the compute the probability over, the output does not have this axis anymore

    Exammple:
        if probs.shape = (N, C) and axis = 1 then entropies.shape = (N, )

    Returns:
        The entropy of the prediction vectors, shape: probs.shape, but without the specified axis
    """
    zeros = np.zeros_like(probs)
    log_probs = np.log2(probs, out=zeros, where=probs > 0)
    entropies = -1 * np.sum(probs * log_probs, axis=axis)
    return entropies

def _margin_largest_secondlargest(probs: np.ndarray) -> np.ndarray:
    """Computes the margin of a probability matrix

        Args:
            probs:
                A probability matrix of shape (N, M)

        Exammple:
            if probs.shape = (N, C) then margins.shape = (N, )

        Returns:
            The margin of the prediction vectors
        """
    sorted_probs = np.partition(probs, -2, axis=1)
    margins = sorted_probs[:, -1] - sorted_probs[:, -2]
    return margins


class ScorerClassification(Scorer):
    """Class to compute active learning scores from the model_output of a classification task.

    Currently supports the following scorers:

        The following three uncertainty scores are taken from [1] and also explained in [2].
        They all have in common, that the score is highest if all classes have the
        same confidence and are 0 if the model assigns 100% probability to a single class.
        The differ in the number of class confidences they take into account.

        - [1]: http://burrsettles.com/pub/settles.activelearning.pdf, Section 3.1, page 12f
        - [2]: https://towardsdatascience.com/uncertainty-sampling-cheatsheet-ec57bc067c0b

        uncertainty_least_confidence:
            This score is 1 - the highest confidence prediction. It is high
            when the confidence about the most probable class is low.

        uncertainty_margin:
            This score is 1 - the margin between the highest conficence
            and second highest confidence prediction. It is high when the model
            cannot decide between the two most probable classes.

        uncertainty_entropy:
            This scorer computes the entropy of the prediction. The confidences
            for all classes are considered to compute the entropy of a sample.

    Attributes:
        model_output:
            Predictions of shape N x C where N is the number of unlabeled samples
            and C is the number of classes in the classification task. Must be
            normalized such that the sum over each row is 1.
            The order of the predictions must be the one specified by
            ActiveLearningAgent.unlabeled_set.

    Examples:
        >>> # example with three unlabeled samples
        >>> al_agent.unlabeled_set
        >>> > ['img0.jpg', 'img1.jpg', 'img2.jpg']
        >>> predictions = np.array(
        >>>     [
        >>>          [0.1, 0.9], # predictions for img0.jpg
        >>>          [0.3, 0.7], # predictions for img1.jpg
        >>>          [0.8, 0.2], # predictions for img2.jpg
        >>>     ] 
        >>> )
        >>> np.sum(predictions, axis=1)
        >>> > array([1., 1., 1.])
        >>> scorer = ScorerClassification(predictions)

    """
    def __init__(self, model_output: Union[np.ndarray, List[List[float]]]):
        if not isinstance(model_output, np.ndarray):
            model_output = np.array(model_output)

        self.model_output = self.ensure_valid_model_output(model_output)

    def ensure_valid_model_output(self, model_output: np.ndarray) -> np.ndarray:
        if len(model_output) == 0:
            return model_output
        if len(model_output.shape) != 2:
            raise ValueError("ScorerClassification model_output must be a 2-dimensional array")
        if model_output.shape[1] == 0:
            raise ValueError("ScorerClassification model_output must not have an empty dimension 1")
        if model_output.shape[1] == 1:
            # assuming a binary classification problem with
            # the model_output denoting the probability of the first class
            model_output = np.concatenate([model_output, 1-model_output], axis=1)
        return model_output

    @classmethod
    def score_names(cls) -> List[str]:
        """Returns the names of the calculated active learning scores
        """
        score_names = list(cls(model_output=[[0.5, 0.5]]).calculate_scores().keys())
        return score_names

    def calculate_scores(self, normalize_to_0_1: bool = True) -> Dict[str, np.ndarray]:
        """Calculates and returns the active learning scores.

        Args:
            normalize_to_0_1:
                If this is true, each score is normalized to have a
                theoretical minimum of 0 and a theoretical maximum of 1.

        Returns:
            A dictionary mapping from the score name (as string)
            to the scores (as a single-dimensional numpy array).
        """
        if len(self.model_output) == 0:
            return {score_name: np.array([]) for score_name in self.score_names()}

        scores_with_names = [
            self._get_scores_uncertainty_least_confidence(),
            self._get_scores_uncertainty_margin(),
            self._get_scores_uncertainty_entropy()
        ]

        scores = dict()
        for score, score_name in scores_with_names:
            score = np.nan_to_num(score)
            scores[score_name] = score

        if normalize_to_0_1:
            scores = self.normalize_scores_0_1(scores)

        return scores

    def normalize_scores_0_1(self, scores: Dict[str, np.ndarray]) -> Dict[str, np.ndarray]:
        num_classes = self.model_output.shape[1]
        model_output_very_sure = np.zeros(shape=(1,num_classes))
        model_output_very_sure[0, 0] = 1
        model_output_very_unsure = np.ones_like(model_output_very_sure)/num_classes

        scores_minimum = ScorerClassification(model_output_very_sure).calculate_scores(normalize_to_0_1=False)
        scores_maximum = ScorerClassification(model_output_very_unsure).calculate_scores(normalize_to_0_1=False)

        for score_name in scores.keys():
            interp_xp = [float(scores_minimum[score_name]), float(scores_maximum[score_name])]
            interp_fp = [0, 1]
            scores[score_name] = np.interp(scores[score_name], interp_xp, interp_fp)

        return scores

    def _get_scores_uncertainty_least_confidence(self):
        scores = 1 - np.max(self.model_output, axis=1)
        return scores, "uncertainty_least_confidence"

    def _get_scores_uncertainty_margin(self):
        scores = 1 - _margin_largest_secondlargest(self.model_output)
        return scores, "uncertainty_margin"

    def _get_scores_uncertainty_entropy(self):
        scores = _entropy(self.model_output, axis=1)
        return scores, "uncertainty_entropy"



from typing import List

from lightly.openapi_generated.swagger_client.models.shared_access_config_create_request import SharedAccessConfigCreateRequest
from lightly.openapi_generated.swagger_client.models.shared_access_config_data import SharedAccessConfigData
from lightly.openapi_generated.swagger_client.models.shared_access_type import SharedAccessType


class _CollaborationMixin:

    def share_dataset_only_with(self, dataset_id: str, user_emails: List[str]):
        """Shares dataset with a list of users

        This method overwrites the list of users that have had access to the dataset
        before. If you want to add someone new to the list make sure you get the
        list of users with access beforehand and add them as well.

        Args:
          dataset_id:
            Identifier of dataset
          user_emails:  
            List of email addresses of users to grant write permission

        Examples:
          >>> # share a dataset with a user
          >>> client = ApiWorkflowClient(token="MY_AWESOME_TOKEN")
          >>> client.share_dataset_only_with(dataset_id="MY_DATASET_ID", user_emails=["user@something.com"])
          >>>
          >>> # share dataset with a user while keep sharing it with previous users
          >>> client = ApiWorkflowClient(token="MY_AWESOME_TOKEN")
          >>> user_emails = client.get_shared_users(dataset_id="MY_DATASET_ID")
          >>> user_emails.append("additional_user2@something.com")
          >>> client.share_dataset_only_with(dataset_id="MY_DATASET_ID", user_emails=user_emails)
          >>>
          >>> # revoke access to all users
          >>> client = ApiWorkflowClient(token="MY_AWESOME_TOKEN")
          >>> client.share_dataset_only_with(dataset_id="MY_DATASET_ID", user_emails=[])
        """
        body = SharedAccessConfigCreateRequest(access_type=SharedAccessType.WRITE, users=user_emails)
        self._collaboration_api.create_or_update_shared_access_config_by_dataset_id(body=body, dataset_id=dataset_id)


    def get_shared_users(self, dataset_id: str) -> List[str]:
      """Get list of users that have access to the dataset
      
      Args:
        dataset_id:
          Identifier of dataset
      
      Returns:
        List of email addresses of users that have write access to the dataset

      Examples:
          >>> client = ApiWorkflowClient(token="MY_AWESOME_TOKEN")
          >>> client.get_shared_users(dataset_id="MY_DATASET_ID")
          >>> ["user@something.com"]
      """

      access_configs: List[SharedAccessConfigData] = self._collaboration_api.get_shared_access_configs_by_dataset_id(dataset_id=dataset_id)
      user_emails = []

      # iterate through configs and find first WRITE config
      # we use the same hard rule in the frontend to communicate with the API
      # as we currently only support WRITE access
      for access_config in access_configs:
        if access_config.access_type == SharedAccessType.WRITE:
          user_emails.extend(access_config.users)
          break

      return user_emails



""" Video Dataset """

# Copyright (c) 2020. Lightly AG and its affiliates.
# All Rights Reserved

import os
from typing import List, Tuple, Dict, Any
from fractions import Fraction
import threading
import weakref
import warnings

import numpy as np
from PIL import Image

import torch
import torchvision
from torch.utils.data import DataLoader, Dataset
from torchvision import datasets
from torchvision import io
from tqdm import tqdm

try:
    import av
    AV_AVAILABLE = True
except ImportError:
    AV_AVAILABLE = False

if io._HAS_VIDEO_OPT:
    torchvision.set_video_backend('video_reader')


class VideoError(Exception):
    """Base exception class for errors during video loading."""
    pass


class EmptyVideoError(VideoError):
    """Exception raised when trying to load a frame from an empty video."""
    pass


class FrameShapeError(VideoError):
    """Exception raised when the loaded frame has an unexpected shape."""
    pass


class NonIncreasingTimestampError(VideoError):
    """Exception raised when trying to load a frame that has a timestamp 
    equal or lower than the timestamps of previous frames in the video.
    """
    pass


class UnseekableTimestampError(VideoError):
    """Exception raised when trying to load a frame that has a timestamp which
    cannot be seeked to by the video loader.
    """
    pass


# @guarin 18.02.2022
# VideoLoader and VideoDataset multi-thread and multi-processing infos
# --------------------------------------------------------------------
# The VideoDataset class should be safe to use in multi-thread and 
# multi-processing settings. For the multi-processing setting it is assumed that
# a pytorch DataLoader is used. Multi-threading should not be use with the
# torchvision pyav video packend as pyav seems to be limited to a single thread.
# You will not see any speedups when using it from multiple threads!
# 
# The VideoLoader class is thread safe because it inherits from threading.local.
# When using it within a pytorch DataLoader a new instance should be created 
# in each process when using the torchvision video_reader backend, otherwise
# decoder errors can happen when iterating multiple times over the dataloader.
# This is specific to the video_reader backend and does not happen with the pyav
# backend.
# 
# In the VideoDataset class we avoid sharing VideoLoader instances between 
# workers by tracking the worker accessing the dataset. VideoLoaders are reset 
# if a new worker accesses the dataset. Note that changes to the dataset class 
# by a worker are unique to that worker and not seen by other workers or the 
# main process.

class VideoLoader(threading.local):
    """Implementation of VideoLoader.

    The VideoLoader is a wrapper around the torchvision video interface. With
    the VideoLoader you can read specific frames or the next frames of a video.
    It automatically switches to the `video_loader` backend if available. Reading
    sequential frames is significantly faster since it uses the VideoReader 
    class from torchvision.

    The video loader automatically detects if you read out subsequent frames and
    will use the fast read method if possible. 

    Attributes:
        path:
            Root directory path.
        timestamps:
            Function that loads file at path.
        backend:
            Tuple of allowed extensions.
        transform:
            Function that takes a PIL image and returns transformed version
        target_transform:
            As transform but for targets
        is_valid_file:
            Used to check corrupt files
        eps:
            Small value to account for floating point imprecisions.

    Examples:
        >>> from torchvision import io
        >>>
        >>> # get timestamps
        >>> ts, fps = io.read_video_timestamps('myvideo.mp4', pts_unit = 'sec')
        >>>
        >>> # create a VideoLoader
        >>> video_loader = VideoLoader('myvideo.mp4', ts)
        >>>
        >>> # get frame at specific timestamp
        >>> frame = video_loader.read_frame(ts[21])
        >>>
        >>> # get next frame
        >>> frame = video_loader.read_frame()
    """
    def __init__(
        self,
        path: str,
        timestamps: List[float],
        backend: str = 'video_reader',
        eps: float = 1e-6,
    ):
        self.path = path
        self.timestamps = timestamps
        self.current_index = None
        self.pts_unit='sec'
        self.backend = backend
        self.eps = eps

        has_video_reader = io._HAS_VIDEO_OPT and hasattr(io, 'VideoReader')

        if has_video_reader and self.backend == 'video_reader':
            self.reader = io.VideoReader(path = self.path)
        else:
            self.reader = None

    def read_frame(self, timestamp = None):
        """Reads the next frame or from timestamp.

        If no timestamp is provided this method just returns the next frame from
        the video. This is significantly (up to 10x) faster if the `video_loader` 
        backend is available. If a timestamp is provided we first have to seek
        to the right position and then load the frame.
        
        Args:
            timestamp: 
                Specific timestamp of frame in seconds or None (default: None)

        Returns:
            A PIL Image

        Raises:
            StopIteration:
                If end of video is reached and timestamp is None.
            ValueError: 
                If provided timestamp is not in self.timestamps.
            VideoError:
                If the frame could not be loaded.

        """
        if not self.timestamps:
            raise EmptyVideoError(
                f'Cannot load frame from empty video {self.path}.'
            )

        if timestamp is None:
            # Try to read next frame.
            if self.current_index is None:
                # Beginning of video.
                index = 0
                timestamp = self.timestamps[index]
            elif self.current_index >= len(self.timestamps):
                # Reached end of video.
                raise StopIteration()
            else:
                # Read next frame.
                index = self.current_index + 1
                timestamp = self.timestamps[index]
        elif (
            self.current_index is not None
            and self.current_index + 1 < len(self.timestamps)
            and timestamp == self.timestamps[self.current_index + 1]
        ):
            # Provided timestamp is timestamp of next frame.
            index = self.current_index + 1
        else:
            # Random timestamp, must find corresponding index.
            index = self.timestamps.index(timestamp)

        if self.reader:
            # Only seek if we cannot just call next(self.reader).
            if (
                self.current_index is None and index != 0
                or self.current_index is not None and index != self.current_index + 1
            ):
                self.reader.seek(timestamp)

            # Find next larger timestamp than the one we seek. Used to verify
            # that we did not seek too far in the video and that the correct
            # frame is returned.
            if index + 1 < len(self.timestamps):
                try:
                    next_timestamp = next(
                        ts for ts in self.timestamps[index + 1:] if ts > timestamp
                    )
                except StopIteration:
                    # All timestamps of future frames are smaller.
                    next_timestamp = float('inf')
            else:
                # Want to load last frame in video.
                next_timestamp = float('inf')

            # Load the frame.
            try:
                while True:
                    frame_info = next(self.reader)
                    if frame_info['pts'] < timestamp - self.eps:
                        # Did not read far enough, let's continue reading more 
                        # frames. This can happen due to decreasing timestamps.
                        frame_info = next(self.reader)
                    elif frame_info['pts'] >= next_timestamp:
                        # Accidentally read too far, let's seek back to the 
                        # correct position and exit. This can happen due to 
                        # imprecise seek.
                        self.reader.seek(timestamp)
                        frame_info = next(self.reader)
                        break
                    else:
                        break
            except StopIteration:
                # Accidentally reached the end of the video, let's seek back to
                # the correction position. This can happen due to imprecise seek.
                self.reader.seek(timestamp)
                try:
                    frame_info = next(self.reader)
                except StopIteration as ex:
                    # Seeking to this timestamp simply doesn't work.
                    raise UnseekableTimestampError(
                        f'Cannot seek to frame with timestamp {float(timestamp)} '
                        f'in {self.path}.'
                    ) from ex

            if (
                frame_info['pts'] < timestamp - self.eps
                or frame_info['pts'] >= next_timestamp
            ):
                # We accidentally loaded the wrong frame. This should only 
                # happen if self.reader.seek(timestamp) does not seek to the
                # correct timestamp. In this case there is nothing we can do to
                # load the correct frame and we alert the user that something
                # went wrong.
                warnings.warn(
                    f'Loaded wrong frame in {self.path}! Tried to load frame '
                    f'with index {index} and timestamp {float(timestamp)} but '
                    f'could only find frame with timestamp {frame_info["pts"]}.'
                )

            # Make sure we have the tensor in correct shape (we want H x W x C)
            frame = frame_info['data'].permute(1,2,0)
            self.current_index = index

        else: # fallback on pyav
            frame, _, _ = io.read_video(self.path,
                                        start_pts=timestamp,
                                        end_pts=timestamp,
                                        pts_unit=self.pts_unit)
            self.current_index = index

        if len(frame.shape) < 3:
            raise FrameShapeError(
                f'Loaded frame has unexpected shape {frame.shape}. '
                f'Frames are expected to have 3 dimensions: (H, W, C).'
            )

        # sometimes torchvision returns multiple frames for one timestamp (bug?)
        if len(frame.shape) > 3 and frame.shape[0] > 1:
            frame = frame[0]

        # make sure we return a H x W x C tensor and not (1 x H x W x C)
        if len(frame.shape) == 4:
            frame = frame.squeeze()

        # convert to PIL image
        image = Image.fromarray(frame.numpy())
        return image


class _TimestampFpsFromVideosDataset(Dataset):

    def __init__(self, video_instances: List[str], pts_unit: str):
        self.video_instances = video_instances
        self.pts_unit = pts_unit

    def __len__(self):
        return len(self.video_instances)

    def __getitem__(self, index):
        instance = self.video_instances[index]
        ts, fps = io.read_video_timestamps(instance, pts_unit=self.pts_unit)
        return ts, fps


def _make_dataset(
        directory,
        extensions=None,
        is_valid_file=None,
        pts_unit='sec',
        tqdm_args=None,
        num_workers: int = 0
):
    """Returns a list of all video files, timestamps, and offsets.

    Args:
        directory:
            Root directory path (should not contain subdirectories).
        extensions:
            Tuple of valid extensions.
        is_valid_file:
            Used to find valid files.
        pts_unit:
            Unit of the timestamps.
        tqdm_args:
            arguments to pass to tqdm
        num_workers:
            number of workers to use for multithreading

    Returns:
        A list of video files, timestamps, frame offsets, and fps.

    """

    if tqdm_args is None:
        tqdm_args = {}
    if extensions is None:
        if is_valid_file is None:
            ValueError('Both extensions and is_valid_file cannot be None')
        else:
            _is_valid_file = is_valid_file
    else:
        def is_valid_file_extension(filepath):
            return filepath.lower().endswith(extensions)
        if is_valid_file is None:
            _is_valid_file = is_valid_file_extension
        else:
            def _is_valid_file(filepath):
                return is_valid_file_extension(filepath) and is_valid_file(filepath)

    # find all video instances (no subdirectories)
    video_instances = []

    def on_error(error):
        raise error
    for root, _, files in os.walk(directory, onerror=on_error):

        for fname in files:
            # skip invalid files
            if not _is_valid_file(os.path.join(root, fname)):
                continue

            # keep track of valid files
            path = os.path.join(root, fname)
            video_instances.append(path)

    # define loader to get the timestamps
    num_workers = min(num_workers, len(video_instances))
    if len(video_instances) == 1:
        num_workers = 0
    loader = DataLoader(
        _TimestampFpsFromVideosDataset(video_instances, pts_unit=pts_unit),
        num_workers=num_workers,
        batch_size=None,
        shuffle=False
    )

    # actually load the data
    tqdm_args = dict(tqdm_args)
    tqdm_args.setdefault('unit', ' video')
    tqdm_args.setdefault('desc', 'Counting frames in videos')
    timestamps_fpss = list(tqdm(loader, **tqdm_args))
    timestamps, fpss = zip(*timestamps_fpss)

    # get frame offsets
    frame_counts = [len(ts) for ts in timestamps]
    offsets = [0] + list(np.cumsum(frame_counts[:-1]))

    return video_instances, timestamps, offsets, fpss


def _find_non_increasing_timestamps(
    timestamps: List[Fraction]
    ) -> List[bool]:
    """Finds all non-increasing timestamps.

    Arguments:
        timestamps:
            Video frame timestamps.

    Returns:
        A boolean for each input timestamp which is True if the timestamp is
        non-increasing and False otherwise.

    """
    if len(timestamps) == 0:
        return []
    is_non_increasing = np.zeros(shape=len(timestamps), dtype=bool, )
    max_timestamp = timestamps[0]-1
    for i, timestamp in enumerate(timestamps):
        if timestamp > max_timestamp:
            max_timestamp = timestamp
        else:
            is_non_increasing[i] = True

    return list(is_non_increasing)


class VideoDataset(datasets.VisionDataset):
    """Implementation of a video dataset.

    The VideoDataset allows random reads from a video file without extracting
    all frames beforehand. This is more storage efficient but is slower.

    Attributes:
        root:
            Root directory path.
        extensions:
            Tuple of allowed extensions.
        transform:
            Function that takes a PIL image and returns transformed version
        target_transform:
            As transform but for targets
        is_valid_file:
            Used to check corrupt files
        exception_on_non_increasing_timestamp:
            If True, a NonIncreasingTimestampError is raised when trying to load
            a frame that has a timestamp lower or equal to the timestamps of 
            previous frames in the same video.

    """

    def __init__(self,
                 root,
                 extensions=None,
                 transform=None,
                 target_transform=None,
                 is_valid_file=None,
                 exception_on_non_increasing_timestamp=True,
                 tqdm_args: Dict[str, Any]=None,
                 num_workers: int = 0,
                 ):

        super(VideoDataset, self).__init__(root,
                                           transform=transform,
                                           target_transform=target_transform)

        videos, video_timestamps, offsets, fps = _make_dataset(
            self.root,
            extensions,
            is_valid_file,
            tqdm_args=tqdm_args,
            num_workers=num_workers
        )

        if len(videos) == 0:
            msg = 'Found 0 videos in folder: {}\n'.format(self.root)
            if extensions is not None:
                msg += 'Supported extensions are: {}'.format(
                    ','.join(extensions))
            raise RuntimeError(msg)

        self.extensions = extensions
        self.backend = torchvision.get_video_backend()
        self.exception_on_non_increasing_timestamp = exception_on_non_increasing_timestamp

        self.videos = videos
        self.video_timestamps = video_timestamps
        self._length = sum((
            len(ts) for ts in self.video_timestamps
        ))
        # Boolean value for every timestamp in self.video_timestamps. If True 
        # the timestamp of the frame is non-increasing compared to timestamps of
        # previous frames in the video.
        self.video_timestamps_is_non_increasing = [
            _find_non_increasing_timestamps(timestamps) for timestamps in video_timestamps
        ]

        # offsets[i] indicates the index of the first frame of the i-th video.
        # e.g. for two videos of length 10 and 20, the offsets will be [0, 10].
        self.offsets = offsets
        self.fps = fps

        # Current VideoLoader instance and the corresponding video index. We 
        # only keep track of the last accessed video as this is a good trade-off
        # between speed and memory requirements.
        # See https://github.com/lightly-ai/lightly/pull/702 for details.
        self._video_loader = None
        self._video_index = None

        # Keep unique reference of dataloader worker. We need this to avoid
        # accidentaly sharing VideoLoader instances between workers.
        self._worker_ref = None

        # Lock to prevent multiple threads creating a new VideoLoader at the 
        # same time.
        self._video_loader_lock = threading.Lock()

    def __getitem__(self, index):
        """Returns item at index.

        Finds the video of the frame at index with the help of the frame 
        offsets. Then, loads the frame from the video, applies the transforms,
        and returns the frame along with the index of the video (as target).

        For example, if there are two videos with 10 and 20 frames respectively
        in the input directory:

        Requesting the 5th sample returns the 5th frame from the first video and
        the target indicates the index of the source video which is 0.
        >>> dataset[5]
        >>> > <PIL Image>, 0

        Requesting the 20th sample returns the 10th frame from the second video
        and the target indicates the index of the source video which is 1.
        >>> dataset[20]
        >>> > <PIL Image>, 1

        Args:
            index:
                Index of the sample to retrieve.

        Returns:
            A tuple (sample, target) where target indicates the video index.

        Raises:
            IndexError:
                If index is out of bounds.
            VideoError:
                If the frame at the given index could not be loaded.

        """
        if index < 0 or index >= self.__len__():
            raise IndexError(f'Index {index} is out of bounds for VideoDataset'
                             f' of size {self.__len__()}.')

        # each sample belongs to a video, to load the sample at index, we need
        # to find the video to which the sample belongs and then read the frame
        # from this video on the disk.
        i = len(self.offsets) - 1
        while (self.offsets[i] > index):
            i = i - 1

        timestamp_idx = index - self.offsets[i]

        if (
            self.exception_on_non_increasing_timestamp
            and self.video_timestamps_is_non_increasing[i][timestamp_idx]
        ):
            raise NonIncreasingTimestampError(
                f'Frame {timestamp_idx} of video {self.videos[i]} has '
                f'a timestamp that is equal or lower than timestamps of previous '
                f'frames in the video. Trying to load this frame might result '
                f'in the wrong frame being returned. Set the VideoDataset.exception_on_non_increasing_timestamp'
                f'attribute to False to allow unsafe frame loading.'
            )

        # find and return the frame as PIL image
        frame_timestamp = self.video_timestamps[i][timestamp_idx]
        video_loader = self._get_video_loader(i)
        sample = video_loader.read_frame(frame_timestamp)

        target = i
        if self.transform is not None:
            sample = self.transform(sample)
        if self.target_transform is not None:
            target = self.target_transform(target)

        return sample, target

    def __len__(self):
        """Returns the number of samples (frames) in the dataset.

        This can be precomputed, because self.video_timestamps is only
        set in the __init__
        """
        return self._length

    def get_filename(self, index):
        """Returns a filename for the frame at index.

        The filename is created from the video filename, the frame number, and
        the video format. The frame number will be zero padded to make sure 
        all filenames have the same length and can easily be sorted.
        E.g. when retrieving a sample from the video
        `my_video.mp4` at frame 153, the filename will be:

        >>> my_video-153-mp4.png
    
        Args:
            index:
                Index of the frame to retrieve.

        Returns:
            The filename of the frame as described above.
                
        """
        if index < 0 or index >= self.__len__():
            raise IndexError(f'Index {index} is out of bounds for VideoDataset'
                             f' of size {self.__len__()}.')

        # each sample belongs to a video, to load the sample at index, we need
        # to find the video to which the sample belongs and then read the frame
        # from this video on the disk.
        i = len(self.offsets) - 1
        while (self.offsets[i] > index):
            i = i - 1

        # get filename of the video file
        video = self.videos[i]
        video_name, video_format = self._video_name_format(video)

        # get frame number
        frame_number = index - self.offsets[i]

        n_frames = self._video_frame_count(i)
        zero_padding = len(str(n_frames))

        return self._format_filename(
            video_name=video_name,
            video_format=video_format,
            frame_number=frame_number,
            zero_padding=zero_padding,
        )

    def get_filenames(self) -> List[str]:
        """Returns a list filenames for all frames in the dataset.
        
        """
        filenames = []
        for i, video in enumerate(self.videos):
            video_name, video_format = self._video_name_format(video)
            n_frames = self._video_frame_count(i)

            zero_padding = len(str(n_frames))
            for frame_number in range(n_frames):
                filenames.append(
                    self._format_filename(
                        video_name=video_name,
                        frame_number=frame_number,
                        video_format=video_format,
                        zero_padding=zero_padding,
                    )
                )
        return filenames

    def _video_frame_count(self, video_index: int) -> int:
        """Returns the number of frames in the video with the given index.
        
        """
        if video_index < len(self.offsets) - 1:
            n_frames = self.offsets[video_index+1] - self.offsets[video_index]
        else:
            n_frames = len(self) - self.offsets[video_index]
        return n_frames

    def _video_name_format(self, video_filename: str) -> Tuple[str, str]:
        """Extracts name and format from the filename of the video.

        Returns:
            A (video_name, video_format) tuple where video_name is the filename
            relative to self.root and video_format is the file extension, for 
            example 'mp4'.

        """
        video_filename = os.path.relpath(video_filename, self.root)
        splits = video_filename.split('.')
        video_format = splits[-1]
        video_name = '.'.join(splits[:-1])
        return video_name, video_format

    def _format_filename(
        self,
        video_name: str,
        frame_number: int,
        video_format: str,
        zero_padding: int = 8,
        extension: str = 'png'
    ) -> str:
        return f'{video_name}-{frame_number:0{zero_padding}}-{video_format}.{extension}'

    def _get_video_loader(self, video_index: int) -> VideoLoader:
        """Returns a video loader unique to the current dataloader worker."""
        worker_info = torch.utils.data.get_worker_info()
        if worker_info is not None:
            # Use a weakref instead of worker_info.id as the worker id is reused
            # by different workers across epochs.
            worker_ref = weakref.ref(worker_info)
            if worker_ref != self._worker_ref:
                # This worker has never accessed the dataset before, we have to
                # reset the video loader.
                self._video_loader = None
                self._video_index = None
                self._worker_ref = worker_ref

        with self._video_loader_lock:
            if video_index != self._video_index:
                video = self.videos[video_index]
                timestamps = self.video_timestamps[video_index]
                self._video_loader = VideoLoader(video, timestamps, backend=self.backend)
                self._video_index = video_index

            return self._video_loader



""" Helper modules for benchmarking SSL models """

# Copyright (c) 2020. Lightly AG and its affiliates.
# All Rights Reserved

import torch
import torch.distributed as dist
import torch.nn as nn
import torch.nn.functional as F
from pytorch_lightning import LightningModule
from torch.utils.data import DataLoader

# code for kNN prediction from here:
# https://colab.research.google.com/github/facebookresearch/moco/blob/colab-notebook/colab/moco_cifar10_demo.ipynb


def knn_predict(feature: torch.Tensor,
                feature_bank: torch.Tensor,
                feature_labels: torch.Tensor, 
                num_classes: int,
                knn_k: int=200,
                knn_t: float=0.1) -> torch.Tensor:
    """Run kNN predictions on features based on a feature bank

    This method is commonly used to monitor performance of self-supervised
    learning methods.

    The default parameters are the ones
    used in https://arxiv.org/pdf/1805.01978v1.pdf.

    Args:
        feature: 
            Tensor of shape [N, D] for which you want predictions
        feature_bank: 
            Tensor of a database of features used for kNN
        feature_labels: 
            Labels for the features in our feature_bank
        num_classes: 
            Number of classes (e.g. `10` for CIFAR-10)
        knn_k: 
            Number of k neighbors used for kNN
        knn_t: 
            Temperature parameter to reweights similarities for kNN

    Returns:
        A tensor containing the kNN predictions

    Examples:
        >>> images, targets, _ = batch
        >>> feature = backbone(images).squeeze()
        >>> # we recommend to normalize the features
        >>> feature = F.normalize(feature, dim=1)
        >>> pred_labels = knn_predict(
        >>>     feature,
        >>>     feature_bank,
        >>>     targets_bank,
        >>>     num_classes=10,
        >>> )
    """

    # compute cos similarity between each feature vector and feature bank ---> [B, N]
    sim_matrix = torch.mm(feature, feature_bank)
    # [B, K]
    sim_weight, sim_indices = sim_matrix.topk(k=knn_k, dim=-1)
    # [B, K]
    sim_labels = torch.gather(feature_labels.expand(
        feature.size(0), -1), dim=-1, index=sim_indices)
    # we do a reweighting of the similarities
    sim_weight = (sim_weight / knn_t).exp()
    # counts for each class
    one_hot_label = torch.zeros(feature.size(
        0) * knn_k, num_classes, device=sim_labels.device)
    # [B*K, C]
    one_hot_label = one_hot_label.scatter(
        dim=-1, index=sim_labels.view(-1, 1), value=1.0)
    # weighted score ---> [B, C]
    pred_scores = torch.sum(one_hot_label.view(feature.size(
        0), -1, num_classes) * sim_weight.unsqueeze(dim=-1), dim=1)
    pred_labels = pred_scores.argsort(dim=-1, descending=True)
    return pred_labels


class BenchmarkModule(LightningModule):
    """A PyTorch Lightning Module for automated kNN callback

    At the end of every training epoch we create a feature bank by feeding the
    `dataloader_kNN` passed to the module through the backbone.
    At every validation step we predict features on the validation data.
    After all predictions on validation data (validation_epoch_end) we evaluate
    the predictions on a kNN classifier on the validation data using the
    feature_bank features from the train data.

    We can access the highest test accuracy during a kNN prediction 
    using the `max_accuracy` attribute.

    Attributes:
        backbone:
            The backbone model used for kNN validation. Make sure that you set the
            backbone when inheriting from `BenchmarkModule`.
        max_accuracy:
            Floating point number between 0.0 and 1.0 representing the maximum
            test accuracy the benchmarked model has achieved.
        dataloader_kNN:
            Dataloader to be used after each training epoch to create feature bank.
        num_classes:
            Number of classes. E.g. for cifar10 we have 10 classes. (default: 10)
        knn_k:
            Number of nearest neighbors for kNN
        knn_t:
            Temperature parameter for kNN

    Examples:
        >>> class SimSiamModel(BenchmarkingModule):
        >>>     def __init__(dataloader_kNN, num_classes):
        >>>         super().__init__(dataloader_kNN, num_classes)
        >>>         resnet = lightly.models.ResNetGenerator('resnet-18')
        >>>         self.backbone = nn.Sequential(
        >>>             *list(resnet.children())[:-1],
        >>>             nn.AdaptiveAvgPool2d(1),
        >>>         )
        >>>         self.resnet_simsiam = 
        >>>             lightly.models.SimSiam(self.backbone, num_ftrs=512)
        >>>         self.criterion = lightly.loss.SymNegCosineSimilarityLoss()
        >>>
        >>>     def forward(self, x):
        >>>         self.resnet_simsiam(x)
        >>>
        >>>     def training_step(self, batch, batch_idx):
        >>>         (x0, x1), _, _ = batch
        >>>         x0, x1 = self.resnet_simsiam(x0, x1)
        >>>         loss = self.criterion(x0, x1)
        >>>         return loss
        >>>     def configure_optimizers(self):
        >>>         optim = torch.optim.SGD(
        >>>             self.resnet_simsiam.parameters(), lr=6e-2, momentum=0.9
        >>>         )
        >>>         return [optim]
        >>>
        >>> model = SimSiamModel(dataloader_train_kNN)
        >>> trainer = pl.Trainer()
        >>> trainer.fit(
        >>>     model,
        >>>     train_dataloader=dataloader_train_ssl,
        >>>     val_dataloaders=dataloader_test
        >>> )
        >>> # you can get the peak accuracy using
        >>> print(model.max_accuracy)

    """

    def __init__(self,
                 dataloader_kNN: DataLoader,
                 num_classes: int,
                 knn_k: int=200,
                 knn_t: float=0.1):
        super().__init__()
        self.backbone = nn.Module()
        self.max_accuracy = 0.0
        self.dataloader_kNN = dataloader_kNN
        self.num_classes = num_classes
        self.knn_k = knn_k
        self.knn_t = knn_t

        # create dummy param to keep track of the device the model is using
        self.dummy_param = nn.Parameter(torch.empty(0))

    def training_epoch_end(self, outputs):
        # update feature bank at the end of each training epoch
        self.backbone.eval()
        self.feature_bank = []
        self.targets_bank = []
        with torch.no_grad():
            for data in self.dataloader_kNN:
                img, target, _ = data
                img = img.to(self.dummy_param.device)
                target = target.to(self.dummy_param.device)
                feature = self.backbone(img).squeeze()
                feature = F.normalize(feature, dim=1)
                self.feature_bank.append(feature)
                self.targets_bank.append(target)
        self.feature_bank = torch.cat(
            self.feature_bank, dim=0).t().contiguous()
        self.targets_bank = torch.cat(
            self.targets_bank, dim=0).t().contiguous()
        self.backbone.train()

    def validation_step(self, batch, batch_idx):
        # we can only do kNN predictions once we have a feature bank
        if hasattr(self, 'feature_bank') and hasattr(self, 'targets_bank'):
            images, targets, _ = batch
            feature = self.backbone(images).squeeze()
            feature = F.normalize(feature, dim=1)
            pred_labels = knn_predict(
                feature,
                self.feature_bank,
                self.targets_bank,
                self.num_classes,
                self.knn_k,
                self.knn_t
            )
            num = images.size()
            top1 = (pred_labels[:, 0] == targets).float().sum()
            return (num, top1)

    def validation_epoch_end(self, outputs):
        device = self.dummy_param.device
        if outputs:
            total_num = torch.Tensor([0]).to(device)
            total_top1 = torch.Tensor([0.]).to(device)
            for (num, top1) in outputs:
                total_num += num[0]
                total_top1 += top1
             
            if dist.is_initialized() and dist.get_world_size() > 1:
                dist.all_reduce(total_num)
                dist.all_reduce(total_top1)

            acc = float(total_top1.item() / total_num.item())
            if acc > self.max_accuracy:
                self.max_accuracy = acc
            self.log('kNN_accuracy', acc * 100.0, prog_bar=True)



# coding: utf-8

"""
    Lightly API

    Lightly.ai enables you to do self-supervised learning in an easy and intuitive way. The lightly.ai OpenAPI spec defines how one can interact with our REST API to unleash the full potential of lightly.ai  # noqa: E501

    OpenAPI spec version: 1.0.0
    Contact: support@lightly.ai
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


import pprint
import re  # noqa: F401

import six

from lightly.openapi_generated.swagger_client.configuration import Configuration


class SampleUpdateRequest(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'file_name': 'str',
        'thumb_name': 'str',
        'exif': 'dict(str, object)',
        'meta_data': 'SampleMetaData',
        'custom_meta_data': 'CustomSampleMetaData'
    }

    attribute_map = {
        'file_name': 'fileName',
        'thumb_name': 'thumbName',
        'exif': 'exif',
        'meta_data': 'metaData',
        'custom_meta_data': 'customMetaData'
    }

    def __init__(self, file_name=None, thumb_name=None, exif=None, meta_data=None, custom_meta_data=None, _configuration=None):  # noqa: E501
        """SampleUpdateRequest - a model defined in Swagger"""  # noqa: E501
        if _configuration is None:
            _configuration = Configuration()
        self._configuration = _configuration

        self._file_name = None
        self._thumb_name = None
        self._exif = None
        self._meta_data = None
        self._custom_meta_data = None
        self.discriminator = None

        if file_name is not None:
            self.file_name = file_name
        if thumb_name is not None:
            self.thumb_name = thumb_name
        if exif is not None:
            self.exif = exif
        if meta_data is not None:
            self.meta_data = meta_data
        if custom_meta_data is not None:
            self.custom_meta_data = custom_meta_data

    @property
    def file_name(self):
        """Gets the file_name of this SampleUpdateRequest.  # noqa: E501


        :return: The file_name of this SampleUpdateRequest.  # noqa: E501
        :rtype: str
        """
        return self._file_name

    @file_name.setter
    def file_name(self, file_name):
        """Sets the file_name of this SampleUpdateRequest.


        :param file_name: The file_name of this SampleUpdateRequest.  # noqa: E501
        :type: str
        """

        self._file_name = file_name

    @property
    def thumb_name(self):
        """Gets the thumb_name of this SampleUpdateRequest.  # noqa: E501


        :return: The thumb_name of this SampleUpdateRequest.  # noqa: E501
        :rtype: str
        """
        return self._thumb_name

    @thumb_name.setter
    def thumb_name(self, thumb_name):
        """Sets the thumb_name of this SampleUpdateRequest.


        :param thumb_name: The thumb_name of this SampleUpdateRequest.  # noqa: E501
        :type: str
        """

        self._thumb_name = thumb_name

    @property
    def exif(self):
        """Gets the exif of this SampleUpdateRequest.  # noqa: E501


        :return: The exif of this SampleUpdateRequest.  # noqa: E501
        :rtype: dict(str, object)
        """
        return self._exif

    @exif.setter
    def exif(self, exif):
        """Sets the exif of this SampleUpdateRequest.


        :param exif: The exif of this SampleUpdateRequest.  # noqa: E501
        :type: dict(str, object)
        """

        self._exif = exif

    @property
    def meta_data(self):
        """Gets the meta_data of this SampleUpdateRequest.  # noqa: E501


        :return: The meta_data of this SampleUpdateRequest.  # noqa: E501
        :rtype: SampleMetaData
        """
        return self._meta_data

    @meta_data.setter
    def meta_data(self, meta_data):
        """Sets the meta_data of this SampleUpdateRequest.


        :param meta_data: The meta_data of this SampleUpdateRequest.  # noqa: E501
        :type: SampleMetaData
        """

        self._meta_data = meta_data

    @property
    def custom_meta_data(self):
        """Gets the custom_meta_data of this SampleUpdateRequest.  # noqa: E501


        :return: The custom_meta_data of this SampleUpdateRequest.  # noqa: E501
        :rtype: CustomSampleMetaData
        """
        return self._custom_meta_data

    @custom_meta_data.setter
    def custom_meta_data(self, custom_meta_data):
        """Sets the custom_meta_data of this SampleUpdateRequest.


        :param custom_meta_data: The custom_meta_data of this SampleUpdateRequest.  # noqa: E501
        :type: CustomSampleMetaData
        """

        self._custom_meta_data = custom_meta_data

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(SampleUpdateRequest, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, SampleUpdateRequest):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, SampleUpdateRequest):
            return True

        return self.to_dict() != other.to_dict()



from typing import List, Dict

import numpy as np

from lightly.active_learning.scorers import Scorer
from lightly.active_learning.utils.keypoint_predictions import \
    KeypointPrediction


def _mean_uncertainty(
        model_output: List[KeypointPrediction]) -> np.ndarray:
    """Score which prefers samples with low confidence score.
    
    The uncertainty score per image is 1 minus the mean confidence
    score of all its keypoints.

    Args:
        model_output:
            Predictions of the model of length N.

    Returns:
        Numpy array of length N with the computed scores.

    """
    scores = []
    for keypoint_prediction in model_output:
        confidences_image = []
        for keypoint_instance_prediction in keypoint_prediction.keypoint_instance_predictions:
            confidences_instance = keypoint_instance_prediction.get_confidences()
            if len(confidences_instance) > 0:
                conf = np.mean(confidences_instance)
                confidences_image.append(conf)
        if len(confidences_image) > 0:
            score = 1. - np.mean(confidences_image)
            scores.append(score)
        else:
            scores.append(0)
    return np.asarray(scores)


class ScorerKeypointDetection(Scorer):
    """Class to compute active learning scores from the model_output of a keypoint detection task.

    Currently supports the following scorers:

        `mean_uncertainty`:
            This scorer uses model predictions to focus more on images which
            have a low mean confidence of the predicted keypoints.
            Use this scorer if you want scenes
            where the model is unsure about the locations of the keypoints.

    Attributes:
        model_output:
            List of KeypointDetectionOutput predicted by the model for several images.


    Examples:
        >>> predictions_over_images = [[{
        >>>     'keypoints': [123., 456., 0.1, 565., 32., 0.2]
        >>> }, {
        >>>     'keypoints': [432., 34., 0.1, 43., 34., 0.3]}
        >>> ], [{
        >>>     'keypoints': [123., 456., 0.1, 565., 32., 0.2])
        >>> }],
        >>> ]
        >>> model_output = []
        >>> for predictions_one_image in predictions_over_images:
        >>>     keypoint_detections = []
        >>>     for prediction in predictions_one_image:
        >>>         keypoints = prediction['keypoints']
        >>>         keypoint_detection = KeypointInstancePrediction(keypoints)
        >>>         keypoint_detections.append(keypoint_detection)
        >>>     output = KeypointPrediction(keypoint_detections)
        >>>     model_output.append(output)
        >>> scorer = ScorerKeypointDetection(model_output)
        >>> scores = scorer.calculate_scores()

    """

    def __init__(self, model_output: List[KeypointPrediction]):
        self.model_output = model_output

    def calculate_scores(self) -> Dict[str, np.ndarray]:
        """Calculates and returns active learning scores in a dictionary.
        """
        # add classification scores
        scores = dict()
        scores['mean_uncertainty'] = _mean_uncertainty(self.model_output)
        return scores

    @classmethod
    def score_names(cls) -> List[str]:
        """Returns the names of the calculated active learning scores
        """
        scorer = cls(model_output=[])
        score_names = list(scorer.calculate_scores().keys())
        return score_names



# coding: utf-8

"""
    Lightly API

    Lightly.ai enables you to do self-supervised learning in an easy and intuitive way. The lightly.ai OpenAPI spec defines how one can interact with our REST API to unleash the full potential of lightly.ai  # noqa: E501

    OpenAPI spec version: 1.0.0
    Contact: support@lightly.ai
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


import pprint
import re  # noqa: F401

import six

from lightly.openapi_generated.swagger_client.configuration import Configuration


class DockerWorkerConfigV2LightlyModel(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'name': 'LightlyModelV2',
        'out_dim': 'int',
        'num_ftrs': 'int',
        'width': 'int'
    }

    attribute_map = {
        'name': 'name',
        'out_dim': 'outDim',
        'num_ftrs': 'numFtrs',
        'width': 'width'
    }

    def __init__(self, name=None, out_dim=None, num_ftrs=None, width=None, _configuration=None):  # noqa: E501
        """DockerWorkerConfigV2LightlyModel - a model defined in Swagger"""  # noqa: E501
        if _configuration is None:
            _configuration = Configuration()
        self._configuration = _configuration

        self._name = None
        self._out_dim = None
        self._num_ftrs = None
        self._width = None
        self.discriminator = None

        if name is not None:
            self.name = name
        if out_dim is not None:
            self.out_dim = out_dim
        if num_ftrs is not None:
            self.num_ftrs = num_ftrs
        if width is not None:
            self.width = width

    @property
    def name(self):
        """Gets the name of this DockerWorkerConfigV2LightlyModel.  # noqa: E501


        :return: The name of this DockerWorkerConfigV2LightlyModel.  # noqa: E501
        :rtype: LightlyModelV2
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this DockerWorkerConfigV2LightlyModel.


        :param name: The name of this DockerWorkerConfigV2LightlyModel.  # noqa: E501
        :type: LightlyModelV2
        """

        self._name = name

    @property
    def out_dim(self):
        """Gets the out_dim of this DockerWorkerConfigV2LightlyModel.  # noqa: E501


        :return: The out_dim of this DockerWorkerConfigV2LightlyModel.  # noqa: E501
        :rtype: int
        """
        return self._out_dim

    @out_dim.setter
    def out_dim(self, out_dim):
        """Sets the out_dim of this DockerWorkerConfigV2LightlyModel.


        :param out_dim: The out_dim of this DockerWorkerConfigV2LightlyModel.  # noqa: E501
        :type: int
        """

        self._out_dim = out_dim

    @property
    def num_ftrs(self):
        """Gets the num_ftrs of this DockerWorkerConfigV2LightlyModel.  # noqa: E501


        :return: The num_ftrs of this DockerWorkerConfigV2LightlyModel.  # noqa: E501
        :rtype: int
        """
        return self._num_ftrs

    @num_ftrs.setter
    def num_ftrs(self, num_ftrs):
        """Sets the num_ftrs of this DockerWorkerConfigV2LightlyModel.


        :param num_ftrs: The num_ftrs of this DockerWorkerConfigV2LightlyModel.  # noqa: E501
        :type: int
        """

        self._num_ftrs = num_ftrs

    @property
    def width(self):
        """Gets the width of this DockerWorkerConfigV2LightlyModel.  # noqa: E501


        :return: The width of this DockerWorkerConfigV2LightlyModel.  # noqa: E501
        :rtype: int
        """
        return self._width

    @width.setter
    def width(self, width):
        """Sets the width of this DockerWorkerConfigV2LightlyModel.


        :param width: The width of this DockerWorkerConfigV2LightlyModel.  # noqa: E501
        :type: int
        """

        self._width = width

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(DockerWorkerConfigV2LightlyModel, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, DockerWorkerConfigV2LightlyModel):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, DockerWorkerConfigV2LightlyModel):
            return True

        return self.to_dict() != other.to_dict()



""" Embedding Strategies """

# Copyright (c) 2020. Lightly AG and its affiliates.
# All Rights Reserved

import time
from typing import List, Union, Tuple

import numpy as np
import torch
import lightly
from lightly.embedding._base import BaseEmbedding
from tqdm import tqdm

from lightly.utils.reordering import sort_items_by_keys

if lightly._is_prefetch_generator_available():
    from prefetch_generator import BackgroundGenerator


class SelfSupervisedEmbedding(BaseEmbedding):
    """Implementation of self-supervised embedding models.

    Implements an embedding strategy based on self-supervised learning. A
    model backbone, self-supervised criterion, optimizer, and dataloader are
    passed to the constructor. The embedding itself is a pytorch-lightning
    module.

    The implementation is based on contrastive learning.

    * SimCLR: https://arxiv.org/abs/2002.05709
    * MoCo: https://arxiv.org/abs/1911.05722
    * SimSiam: https://arxiv.org/abs/2011.10566

    Attributes:
        model:
            A backbone convolutional network with a projection head.
        criterion:
            A contrastive loss function.
        optimizer:
            A PyTorch optimizer.
        dataloader:
            A torchvision dataloader.
        scheduler:
            A PyTorch learning rate scheduler.

    Examples:
        >>> # define a model, criterion, optimizer, and dataloader above
        >>> import lightly.embedding as embedding
        >>> encoder = SelfSupervisedEmbedding(
        >>>     model,
        >>>     criterion,
        >>>     optimizer,
        >>>     dataloader,
        >>> )
        >>> # train the self-supervised embedding with default settings
        >>> encoder.train_embedding()
        >>> # pass pytorch-lightning trainer arguments as kwargs
        >>> encoder.train_embedding(max_epochs=10)

    """

    def __init__(
        self,
        model: torch.nn.Module,
        criterion: torch.nn.Module,
        optimizer: torch.optim.Optimizer,
        dataloader: torch.utils.data.DataLoader,
        scheduler=None,
    ):

        super(SelfSupervisedEmbedding, self).__init__(
            model, criterion, optimizer, dataloader, scheduler
        )

    def embed(self,
              dataloader: torch.utils.data.DataLoader,
              device: torch.device = None
    ) -> Tuple[np.ndarray, List[int], List[str]]:
        """Embeds images in a vector space.

        Args:
            dataloader:
                A PyTorch dataloader.
            device:
                Selected device (`cpu`, `cuda`, see PyTorch documentation)

        Returns:
            Tuple of (embeddings, labels, filenames) ordered by the
            samples in the dataset of the dataloader.
                embeddings:
                    Embedding of shape (n_samples, embedding_feature_size).
                    One embedding for each sample.
                labels:
                    Labels of shape (n_samples, ).
                filenames:
                    The filenames from dataloader.dataset.get_filenames().


        Examples:
            >>> # embed images in vector space
            >>> embeddings, labels, fnames = encoder.embed(dataloader)

        """

        self.model.eval()
        embeddings, labels, filenames = None, None, []

        dataset = dataloader.dataset
        if lightly._is_prefetch_generator_available():
            dataloader = BackgroundGenerator(dataloader, max_prefetch=3)
        
        pbar = tqdm(
            total=len(dataset),
            unit='imgs'
        )

        efficiency = 0.0
        embeddings = []
        labels = []
        with torch.no_grad():

            start_timepoint = time.time()
            for (image_batch, label_batch, filename_batch) in dataloader:

                batch_size = image_batch.shape[0]

                # the following 2 lines are needed to prevent a file handler leak,
                # see https://github.com/lightly-ai/lightly/pull/676
                image_batch = image_batch.to(device)
                label_batch = label_batch.clone()

                filenames += [*filename_batch]

                prepared_timepoint = time.time()

                embedding_batch = self.model.backbone(image_batch)
                embedding_batch = embedding_batch.detach().reshape(batch_size, -1)

                embeddings.append(embedding_batch)
                labels.append(label_batch)

                finished_timepoint = time.time()

                data_loading_time = prepared_timepoint - start_timepoint
                inference_time = finished_timepoint - prepared_timepoint
                total_batch_time = data_loading_time + inference_time

                efficiency = inference_time / total_batch_time
                pbar.set_description("Compute efficiency: {:.2f}".format(efficiency))
                start_timepoint = time.time()

                pbar.update(batch_size)

            embeddings = torch.cat(embeddings, 0)
            labels = torch.cat(labels, 0)

            embeddings = embeddings.cpu().numpy()
            labels = labels.cpu().numpy()

        sorted_filenames = dataset.get_filenames()
        sorted_embeddings = sort_items_by_keys(
            filenames, embeddings, sorted_filenames
        )
        sorted_labels = sort_items_by_keys(
            filenames, labels, sorted_filenames
        )
        embeddings = np.stack(sorted_embeddings)
        labels = np.stack(sorted_labels).tolist()

        return embeddings, labels, sorted_filenames



""" Collection of Utils for Active Learning """

# Copyright (c) 2020. Lightly AG and its affiliates.
# All Rights Reserved

from lightly.active_learning.utils.bounding_box import BoundingBox
from lightly.active_learning.utils.object_detection_output import ObjectDetectionOutput


""" Gaussian Blur """

# Copyright (c) 2020. Lightly AG and its affiliates.
# All Rights Reserved

import numpy as np
from PIL import ImageFilter
from typing import Tuple


class GaussianBlur(object):
    """Implementation of random Gaussian blur.

    Utilizes the built-in ImageFilter method from PIL to apply a Gaussian 
    blur to the input image with a certain probability. The blur is further
    randomized by sampling uniformly the values of the standard deviation of
    the Gaussian kernel.

    Attributes:
        kernel_size:
            Mean kernel size for the Gaussian blur.
        prob:
            Probability with which the blur is applied.
        scale:
            Old unused parameters kept for compatibility
        sigmas:
            Tuple of min and max value from which the std of the gaussian kernel is sampled

    """

    def __init__(self, kernel_size: float, prob: float = 0.5,
                 scale: float = 0.2, sigmas: Tuple[float, float] = (.2, 2)):
        self.prob = prob
        self.sigmas = sigmas

    def __call__(self, sample):
        """Blurs the image with a given probability.

        Args:
            sample:
                PIL image to which blur will be applied.
        
        Returns:
            Blurred image or original image.
        """
        prob = np.random.random_sample()
        if prob < self.prob:
            # choose randomized std for Gaussian filtering
            sigma = np.random.uniform(self.sigmas[0], self.sigmas[1])
            return sample.filter(ImageFilter.GaussianBlur(radius=sigma))
        # return original image
        return sample



""" Helper Functions """

# Copyright (c) 2020. Lightly AG and its affiliates.
# All Rights Reserved

import os
from typing import List, Set, Optional, Callable, Dict, Any

from torchvision import datasets

from lightly.data._image import DatasetFolder

try:
    from lightly.data._video import VideoDataset
    VIDEO_DATASET_AVAILABLE = True
except Exception as e:
    VIDEO_DATASET_AVAILABLE = False
    VIDEO_DATASET_ERRORMSG = e


IMG_EXTENSIONS = ('.jpg', '.jpeg', '.png', '.ppm', '.bmp',
                  '.pgm', '.tif', '.tiff', '.webp')

VIDEO_EXTENSIONS = ('.mp4', '.mov', '.avi', '.mpg',
                    '.hevc', '.m4v', '.webm', '.mpeg')


def _dir_contains_videos(root: str, extensions: tuple):
    """Checks whether directory contains video files.

    Args:
        root: Root directory path.

    Returns:
        True if root contains video files.

    """
    with os.scandir(root) as scan_dir:
        return any(f.name.lower().endswith(extensions) for f in scan_dir)


def _contains_videos(root: str, extensions: tuple):
    """Checks whether directory or any subdirectory contains video files.

    Iterates over all subdirectories of "root" recursively and returns True
    if any of the subdirectories contains a file with a VIDEO_EXTENSION.

    Args:
        root: Root directory path.

    Returns:
        True if "root" or any subdir contains video files.

    """
    for subdir, _, _ in os.walk(root):
        if _dir_contains_videos(subdir, extensions):
            return True
    return False


def _is_lightly_output_dir(dirname: str):
    """Checks whether the directory is a lightly_output directory.

    Args:
        dirname: Directory to check.

    Returns:
        True if dirname is "lightly_outputs" else false.

    """
    return 'lightly_outputs' in dirname


def _contains_subdirs(root: str):
    """Checks whether directory contains subdirectories.

    Args:
        root: Root directory path.

    Returns:
        True if root contains subdirectories else false.

    """
    with os.scandir(root) as scan_dir:
        return any(not _is_lightly_output_dir(f.name) for f in scan_dir \
            if f.is_dir())


def _load_dataset_from_folder(
        root: str, transform,
        is_valid_file: Optional[Callable[[str], bool]] = None,
        tqdm_args: Dict[str, Any] = None,
        num_workers_video_frame_counting: int = 0
):
    """Initializes dataset from folder.

    Args:
        root: (str) Root directory path
        transform: (torchvision.transforms.Compose) image transformations

    Returns:
        Dataset consisting of images/videos in the root directory.

    Raises:
        ValueError: If the specified dataset doesn't exist

    """
    if not os.path.exists(root):
        raise ValueError(f'The input directory {root} does not exist!')

    # if there is a video in the input directory but we do not have
    # the right dependencies, raise a ValueError
    contains_videos = _contains_videos(root, VIDEO_EXTENSIONS)
    if contains_videos and not VIDEO_DATASET_AVAILABLE:
        raise ValueError(f'The input directory {root} contains videos '
                         'but the VideoDataset is not available. \n'
                         'Make sure you have installed the right '
                         'dependencies. The error from the imported '
                         f'module was: {VIDEO_DATASET_ERRORMSG}')

    if contains_videos:
        # root contains videos -> create a video dataset
        dataset = VideoDataset(
            root,
            extensions=VIDEO_EXTENSIONS,
            transform=transform,
            is_valid_file=is_valid_file,
            tqdm_args=tqdm_args,
            num_workers=num_workers_video_frame_counting
        )
    elif _contains_subdirs(root):
        # root contains subdirectories -> create an image folder dataset
        dataset = datasets.ImageFolder(root,
                                       transform=transform,
                                       is_valid_file=is_valid_file
                                       )
    else:
        # root contains plain images -> create a folder dataset
        dataset = DatasetFolder(root,
                                extensions=IMG_EXTENSIONS,
                                transform=transform,
                                is_valid_file=is_valid_file
                                )

    return dataset



""" Command-Line Interface Helpers """

# Copyright (c) 2020. Lightly AG and its affiliates.
# All Rights Reserved
import copy
import os
import warnings
from typing import Type

import torch
import hydra
from hydra import utils
from torch import nn as nn
from torch.utils.hipify.hipify_python import bcolors

from lightly.utils import version_compare

from lightly.cli._cli_simclr import _SimCLR
from lightly.embedding import SelfSupervisedEmbedding

from lightly.models import ZOO as model_zoo, ResNetGenerator
from lightly.models.batchnorm import get_norm_layer


def _custom_formatwarning(msg, *args, **kwargs):
    # ignore everything except the message
    return f"{bcolors.WARNING}{msg}{bcolors.WARNING}\n"


def print_as_warning(message: str, warning_class: Type[Warning] = UserWarning):
    old_format = copy.copy(warnings.formatwarning)

    warnings.formatwarning = _custom_formatwarning
    warnings.warn(message, warning_class)

    warnings.formatwarning = old_format


def cpu_count():
    """Returns the number of CPUs which are present in the system.

    This number is not equivalent to the number of available CPUs to the process.

    """
    return os.cpu_count()


def fix_input_path(path):
    """Fix broken relative paths.

    """
    if not os.path.isabs(path):
        path = utils.to_absolute_path(path)
    return path


def fix_hydra_arguments(config_path: str = 'config', config_name: str = 'config'):
    """Helper to make hydra arugments adaptive to installed hydra version
    
    Hydra introduced the `version_base` argument in version 1.2.0
    We use this helper to provide backwards compatibility to older hydra verisons.    
    """

    hydra_args = {'config_path': config_path, 'config_name': config_name}

    try:
        if version_compare(hydra.__version__, '1.1.2') > 0:
            hydra_args['version_base'] = '1.1'
    except ValueError:
        pass
    
    return hydra_args


def is_url(checkpoint):
    """Check whether the checkpoint is a url or not.

    """
    is_url = ('https://storage.googleapis.com' in checkpoint)
    return is_url


def get_ptmodel_from_config(model):
    """Get a pre-trained model from the lightly model zoo.

    """
    key = model['name']
    key += '/simclr'
    key += '/d' + str(model['num_ftrs'])
    key += '/w' + str(float(model['width']))

    if key in model_zoo.keys():
        return model_zoo[key], key
    else:
        return '', key


def load_state_dict_from_url(url, map_location=None):
    """Try to load the checkopint from the given url.

    """
    try:
        state_dict = torch.hub.load_state_dict_from_url(
            url, map_location=map_location
        )
        return state_dict
    except Exception:
        print('Not able to load state dict from %s' % (url))
        print('Retrying with http:// prefix')
    try:
        url = url.replace('https', 'http')
        state_dict = torch.hub.load_state_dict_from_url(
            url, map_location=map_location
        )
        return state_dict
    except Exception:
        print('Not able to load state dict from %s' % (url))

    # in this case downloading the pre-trained model was not possible
    # notify the user and return
    return {'state_dict': None}


def _maybe_expand_batchnorm_weights(model_dict, state_dict, num_splits):
    """Expands the weights of the BatchNorm2d to the size of SplitBatchNorm.

    """
    running_mean = 'running_mean'
    running_var = 'running_var'

    for key, item in model_dict.items():
        # not batchnorm -> continue
        if not running_mean in key and not running_var in key:
            continue

        state = state_dict.get(key, None)
        # not in dict -> continue
        if state is None:
            continue
        # same shape -> continue
        if item.shape == state.shape:
            continue

        # found running mean or running var with different shapes
        state_dict[key] = state.repeat(num_splits)

    return state_dict


def _filter_state_dict(state_dict, remove_model_prefix_offset: int = 1):
    """Makes the state_dict compatible with the model.
    
    Prevents unexpected key error when loading PyTorch-Lightning checkpoints.
    Allows backwards compatability to checkpoints before v1.0.6.

    """

    prev_backbone = 'features'
    curr_backbone = 'backbone'

    new_state_dict = {}
    for key, item in state_dict.items():
        # remove the "model." prefix from the state dict key
        key_parts = key.split('.')[remove_model_prefix_offset:]
        # with v1.0.6 the backbone of the models will be renamed from
        # "features" to "backbone", ensure compatability with old ckpts
        key_parts = \
            [k if k != prev_backbone else curr_backbone for k in key_parts]

        new_key = '.'.join(key_parts)
        new_state_dict[new_key] = item

    return new_state_dict


def _fix_projection_head_keys(state_dict):
    """Makes the state_dict compatible with the refactored projection heads.

    TODO: Remove once the models are refactored and the old checkpoints were
    replaced! Relevant issue: https://github.com/lightly-ai/lightly/issues/379

    Prevents unexpected key error when loading old checkpoints.
    
    """

    projection_head_identifier = 'projection_head'
    prediction_head_identifier = 'prediction_head'
    projection_head_insert = 'layers'

    new_state_dict = {}
    for key, item in state_dict.items():
        if (projection_head_identifier in key or \
            prediction_head_identifier in key) and \
                projection_head_insert not in key:
            # insert layers if it's not part of the key yet
            key_parts = key.split('.')
            key_parts.insert(1, projection_head_insert)
            new_key = '.'.join(key_parts)
        else:
            new_key = key

        new_state_dict[new_key] = item

    return new_state_dict


def load_from_state_dict(model,
                         state_dict,
                         strict: bool = True,
                         apply_filter: bool = True,
                         num_splits: int = 0):
    """Loads the model weights from the state dictionary.

    """

    # step 1: filter state dict
    if apply_filter:
        state_dict = _filter_state_dict(state_dict)

    state_dict = _fix_projection_head_keys(state_dict)

    # step 2: expand batchnorm weights
    state_dict = \
        _maybe_expand_batchnorm_weights(model.state_dict(), state_dict, num_splits)

    # step 3: load from checkpoint
    model.load_state_dict(state_dict, strict=strict)


def get_model_from_config(cfg, is_cli_call: bool = False) -> SelfSupervisedEmbedding:
    checkpoint = cfg['checkpoint']
    if torch.cuda.is_available():
        device = torch.device('cuda')
    else:
        device = torch.device('cpu')

    if not checkpoint:
        checkpoint, key = get_ptmodel_from_config(cfg['model'])
        if not checkpoint:
            msg = 'Cannot download checkpoint for key {} '.format(key)
            msg += 'because it does not exist!'
            raise RuntimeError(msg)
        state_dict = load_state_dict_from_url(checkpoint, map_location=device)[
            'state_dict'
        ]
    else:
        checkpoint = fix_input_path(checkpoint) if is_cli_call else checkpoint
        state_dict = torch.load(checkpoint, map_location=device)['state_dict']

    # load model
    resnet = ResNetGenerator(cfg['model']['name'], cfg['model']['width'])
    last_conv_channels = list(resnet.children())[-1].in_features
    features = nn.Sequential(
        get_norm_layer(3, 0),
        *list(resnet.children())[:-1],
        nn.Conv2d(last_conv_channels, cfg['model']['num_ftrs'], 1),
        nn.AdaptiveAvgPool2d(1),
    )

    model = _SimCLR(
        features, num_ftrs=cfg['model']['num_ftrs'], out_dim=cfg['model']['out_dim']
    ).to(device)

    if state_dict is not None:
        load_from_state_dict(model, state_dict)

    encoder = SelfSupervisedEmbedding(model, None, None, None)
    return encoder



# coding: utf-8

"""
    Lightly API

    Lightly.ai enables you to do self-supervised learning in an easy and intuitive way. The lightly.ai OpenAPI spec defines how one can interact with our REST API to unleash the full potential of lightly.ai  # noqa: E501

    OpenAPI spec version: 1.0.0
    Contact: support@lightly.ai
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


import pprint
import re  # noqa: F401

import six

from lightly.openapi_generated.swagger_client.configuration import Configuration


class SamplingCreateRequest(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'new_tag_name': 'TagName',
        'method': 'SamplingMethod',
        'config': 'SamplingConfig',
        'preselected_tag_id': 'MongoObjectID',
        'query_tag_id': 'MongoObjectID',
        'score_type': 'ActiveLearningScoreType',
        'row_count': 'float'
    }

    attribute_map = {
        'new_tag_name': 'newTagName',
        'method': 'method',
        'config': 'config',
        'preselected_tag_id': 'preselectedTagId',
        'query_tag_id': 'queryTagId',
        'score_type': 'scoreType',
        'row_count': 'rowCount'
    }

    def __init__(self, new_tag_name=None, method=None, config=None, preselected_tag_id=None, query_tag_id=None, score_type=None, row_count=None, _configuration=None):  # noqa: E501
        """SamplingCreateRequest - a model defined in Swagger"""  # noqa: E501
        if _configuration is None:
            _configuration = Configuration()
        self._configuration = _configuration

        self._new_tag_name = None
        self._method = None
        self._config = None
        self._preselected_tag_id = None
        self._query_tag_id = None
        self._score_type = None
        self._row_count = None
        self.discriminator = None

        self.new_tag_name = new_tag_name
        self.method = method
        self.config = config
        if preselected_tag_id is not None:
            self.preselected_tag_id = preselected_tag_id
        if query_tag_id is not None:
            self.query_tag_id = query_tag_id
        if score_type is not None:
            self.score_type = score_type
        if row_count is not None:
            self.row_count = row_count

    @property
    def new_tag_name(self):
        """Gets the new_tag_name of this SamplingCreateRequest.  # noqa: E501


        :return: The new_tag_name of this SamplingCreateRequest.  # noqa: E501
        :rtype: TagName
        """
        return self._new_tag_name

    @new_tag_name.setter
    def new_tag_name(self, new_tag_name):
        """Sets the new_tag_name of this SamplingCreateRequest.


        :param new_tag_name: The new_tag_name of this SamplingCreateRequest.  # noqa: E501
        :type: TagName
        """
        if self._configuration.client_side_validation and new_tag_name is None:
            raise ValueError("Invalid value for `new_tag_name`, must not be `None`")  # noqa: E501

        self._new_tag_name = new_tag_name

    @property
    def method(self):
        """Gets the method of this SamplingCreateRequest.  # noqa: E501


        :return: The method of this SamplingCreateRequest.  # noqa: E501
        :rtype: SamplingMethod
        """
        return self._method

    @method.setter
    def method(self, method):
        """Sets the method of this SamplingCreateRequest.


        :param method: The method of this SamplingCreateRequest.  # noqa: E501
        :type: SamplingMethod
        """
        if self._configuration.client_side_validation and method is None:
            raise ValueError("Invalid value for `method`, must not be `None`")  # noqa: E501

        self._method = method

    @property
    def config(self):
        """Gets the config of this SamplingCreateRequest.  # noqa: E501


        :return: The config of this SamplingCreateRequest.  # noqa: E501
        :rtype: SamplingConfig
        """
        return self._config

    @config.setter
    def config(self, config):
        """Sets the config of this SamplingCreateRequest.


        :param config: The config of this SamplingCreateRequest.  # noqa: E501
        :type: SamplingConfig
        """
        if self._configuration.client_side_validation and config is None:
            raise ValueError("Invalid value for `config`, must not be `None`")  # noqa: E501

        self._config = config

    @property
    def preselected_tag_id(self):
        """Gets the preselected_tag_id of this SamplingCreateRequest.  # noqa: E501


        :return: The preselected_tag_id of this SamplingCreateRequest.  # noqa: E501
        :rtype: MongoObjectID
        """
        return self._preselected_tag_id

    @preselected_tag_id.setter
    def preselected_tag_id(self, preselected_tag_id):
        """Sets the preselected_tag_id of this SamplingCreateRequest.


        :param preselected_tag_id: The preselected_tag_id of this SamplingCreateRequest.  # noqa: E501
        :type: MongoObjectID
        """

        self._preselected_tag_id = preselected_tag_id

    @property
    def query_tag_id(self):
        """Gets the query_tag_id of this SamplingCreateRequest.  # noqa: E501


        :return: The query_tag_id of this SamplingCreateRequest.  # noqa: E501
        :rtype: MongoObjectID
        """
        return self._query_tag_id

    @query_tag_id.setter
    def query_tag_id(self, query_tag_id):
        """Sets the query_tag_id of this SamplingCreateRequest.


        :param query_tag_id: The query_tag_id of this SamplingCreateRequest.  # noqa: E501
        :type: MongoObjectID
        """

        self._query_tag_id = query_tag_id

    @property
    def score_type(self):
        """Gets the score_type of this SamplingCreateRequest.  # noqa: E501


        :return: The score_type of this SamplingCreateRequest.  # noqa: E501
        :rtype: ActiveLearningScoreType
        """
        return self._score_type

    @score_type.setter
    def score_type(self, score_type):
        """Sets the score_type of this SamplingCreateRequest.


        :param score_type: The score_type of this SamplingCreateRequest.  # noqa: E501
        :type: ActiveLearningScoreType
        """

        self._score_type = score_type

    @property
    def row_count(self):
        """Gets the row_count of this SamplingCreateRequest.  # noqa: E501

        temporary rowCount until the API/DB is aware how many they are..  # noqa: E501

        :return: The row_count of this SamplingCreateRequest.  # noqa: E501
        :rtype: float
        """
        return self._row_count

    @row_count.setter
    def row_count(self, row_count):
        """Sets the row_count of this SamplingCreateRequest.

        temporary rowCount until the API/DB is aware how many they are..  # noqa: E501

        :param row_count: The row_count of this SamplingCreateRequest.  # noqa: E501
        :type: float
        """

        self._row_count = row_count

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(SamplingCreateRequest, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, SamplingCreateRequest):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, SamplingCreateRequest):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Lightly API

    Lightly.ai enables you to do self-supervised learning in an easy and intuitive way. The lightly.ai OpenAPI spec defines how one can interact with our REST API to unleash the full potential of lightly.ai  # noqa: E501

    OpenAPI spec version: 1.0.0
    Contact: support@lightly.ai
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


import pprint
import re  # noqa: F401

import six

from lightly.openapi_generated.swagger_client.configuration import Configuration


class DatasourceConfig(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
    }

    attribute_map = {
    }

    discriminator_value_class_map = {
        
    }

    def __init__(self, _configuration=None):  # noqa: E501
        """DatasourceConfig - a model defined in Swagger"""  # noqa: E501
        if _configuration is None:
            _configuration = Configuration()
        self._configuration = _configuration
        self.discriminator = 'Discriminator{propertyName&#x3D;&#39;type&#39;, mapping&#x3D;{LIGHTLY&#x3D;#/components/schemas/DatasourceConfigLIGHTLY, S3&#x3D;#/components/schemas/DatasourceConfigS3, S3DelegatedAccess&#x3D;#/components/schemas/DatasourceConfigS3DelegatedAccess, GCS&#x3D;#/components/schemas/DatasourceConfigGCS, AZURE&#x3D;#/components/schemas/DatasourceConfigAzure, OBS&#x3D;#/components/schemas/DatasourceConfigOBS, LOCAL&#x3D;#/components/schemas/DatasourceConfigLOCAL}, extensions&#x3D;null}'

    def get_real_child_model(self, data):
        """Returns the real base class specified by the discriminator"""
        discriminator_value = data[self.discriminator].lower()
        return self.discriminator_value_class_map.get(discriminator_value)

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(DatasourceConfig, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, DatasourceConfig):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, DatasourceConfig):
            return True

        return self.to_dict() != other.to_dict()



# -*- coding: utf-8 -*-
"""**Lightly Embed:** Embed images with one command.

This module contains the entrypoint for the **lightly-embed**
command-line interface.
"""

# Copyright (c) 2020. Lightly AG and its affiliates.
# All Rights Reserved

import os
from typing import Union, Tuple, List

import hydra
import numpy as np
import torch
import torchvision
from torch.utils.hipify.hipify_python import bcolors

from lightly.data import LightlyDataset

from lightly.cli._helpers import fix_hydra_arguments

from lightly.utils import save_embeddings

from lightly.cli._helpers import get_model_from_config
from lightly.cli._helpers import fix_input_path
from lightly.cli._helpers import cpu_count


def _embed_cli(cfg, is_cli_call=True) -> \
    Union[
        Tuple[np.ndarray, List[int], List[str]],
        str
    ]:
    """ See embed_cli() for usage documentation

        is_cli_call:
            If True:
                Saves the embeddings as file and returns the filepath.
            If False:
                Returns the embeddings, labels, filenames as tuple.
                Embeddings are of shape (n_samples, embedding_size)
                len(labels) = len(filenames) = n_samples
    """
    input_dir = cfg['input_dir']
    if input_dir and is_cli_call:
        input_dir = fix_input_path(input_dir)

    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False

    if torch.cuda.is_available():
        device = torch.device('cuda')
    else:
        device = torch.device('cpu')

    transform = torchvision.transforms.Compose(
        [
            torchvision.transforms.Resize(
                (cfg['collate']['input_size'], cfg['collate']['input_size'])
            ),
            torchvision.transforms.ToTensor(),
            torchvision.transforms.Normalize(
                mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]
            ),
        ]
    )

    dataset = LightlyDataset(input_dir, transform=transform)

    # disable drop_last and shuffle
    cfg['loader']['drop_last'] = False
    cfg['loader']['shuffle'] = False
    cfg['loader']['batch_size'] = min(cfg['loader']['batch_size'], len(dataset))

    # determine the number of available cores
    if cfg['loader']['num_workers'] < 0:
        cfg['loader']['num_workers'] = cpu_count()

    dataloader = torch.utils.data.DataLoader(dataset, **cfg['loader'])

    encoder = get_model_from_config(cfg, is_cli_call)

    embeddings, labels, filenames = encoder.embed(dataloader, device=device)

    if is_cli_call:
        path = os.path.join(os.getcwd(), 'embeddings.csv')
        save_embeddings(path, embeddings, labels, filenames)
        print(f'Embeddings are stored at {bcolors.OKBLUE}{path}{bcolors.ENDC}')
        os.environ[
            cfg['environment_variable_names']['lightly_last_embedding_path']
        ] = path
        return path

    return embeddings, labels, filenames


@hydra.main(**fix_hydra_arguments(config_path = 'config', config_name = 'config'))
def embed_cli(cfg) -> str:
    """Embed images from the command-line.

    Args:
        cfg:
            The default configs are loaded from the config file.
            To overwrite them please see the section on the config file
            (.config.config.yaml).

    Command-Line Args:
        input_dir:
            Path to the input directory where images are stored.
        checkpoint:
            Path to the checkpoint of a pretrained model. If left
            empty, a pretrained model by lightly is used.

    Returns:
        The path to the created embeddings file.

    Examples:
        >>> # embed images with default settings and a lightly model
        >>> lightly-embed input_dir=data/
        >>>
        >>> # embed images with default settings and a custom checkpoint
        >>> lightly-embed input_dir=data/ checkpoint=my_checkpoint.ckpt
        >>>
        >>> # embed images with custom settings
        >>> lightly-embed input_dir=data/ model.num_ftrs=32

    """
    return _embed_cli(cfg)


def entry():
    embed_cli()



# -*- coding: utf-8 -*-
"""**Lightly Upload:** Upload images to the Lightly platform.

This module contains the entrypoint for the **lightly-upload**
command-line interface.
"""

# Copyright (c) 2020. Lightly AG and its affiliates.
# All Rights Reserved
import csv
import json
import os
import warnings
from datetime import datetime
from typing import Union

import hydra

import torchvision
from torch.utils.hipify.hipify_python import bcolors

from lightly.api.api_workflow_upload_embeddings import \
    EmbeddingDoesNotExistError
    
from lightly.cli._helpers import fix_input_path
from lightly.cli._helpers import print_as_warning
from lightly.cli._helpers import cpu_count
from lightly.cli._helpers import fix_hydra_arguments

from lightly.api.api_workflow_client import ApiWorkflowClient
from lightly.data import LightlyDataset
from lightly.openapi_generated.swagger_client import DatasetCreator

SUCCESS_RETURN_VALUE = "Success"


def _upload_cli(cfg, is_cli_call=True) -> Union[str, None]:
    """

    Returns:
        if no errors were encountered:
            SUCCESS_RETURN_VALUE
        else:
            None

    """
    # TODO (Malte, 19.01,2023): Remove the _upload_cli completely
    print_as_warning(
        message="DeprecationWarning: Uploading via CLI is deprecated and will be removed soon! "
        "Please use the Lightly Worker instead: https://docs.lightly.ai/docs/install-lightly\n",
    )

    input_dir = cfg['input_dir']
    if input_dir and is_cli_call:
        input_dir = fix_input_path(input_dir)

    path_to_embeddings = cfg['embeddings']
    if path_to_embeddings and is_cli_call:
        path_to_embeddings = fix_input_path(path_to_embeddings)

    dataset_id = cfg['dataset_id']
    token = cfg['token']
    new_dataset_name = cfg['new_dataset_name']

    cli_api_args_wrong = False
    if not token:
        print_as_warning('Please specify your access token.')
        cli_api_args_wrong = True

    if dataset_id:
        if new_dataset_name:
            print_as_warning(
                'Please specify either the dataset_id of an existing dataset '
                'or a new_dataset_name, but not both.'
            )
            cli_api_args_wrong = True
        else:
            api_workflow_client = \
                ApiWorkflowClient(token=token, dataset_id=dataset_id, dataset_creator=DatasetCreator.USER_PIP_LIGHTLY_MAGIC)
    else:
        if new_dataset_name:
            api_workflow_client = ApiWorkflowClient(token=token, dataset_creator=DatasetCreator.USER_PIP_LIGHTLY_MAGIC)
            api_workflow_client.create_dataset(dataset_name=new_dataset_name)
        else:
            print_as_warning(
                'Please specify either the dataset_id of an existing dataset '
                'or a new_dataset_name.')
            cli_api_args_wrong = True
    # delete the dataset_id as it might be an empty string
    # Use api_workflow_client.dataset_id instead
    del dataset_id

    if cli_api_args_wrong:
        print_as_warning('For help, try: lightly-upload --help')
        return

    # potentially load custom metadata
    custom_metadata = None
    if cfg['custom_metadata']:
        path_to_custom_metadata = fix_input_path(cfg['custom_metadata'])
        print(
            'Loading custom metadata from '
            f'{bcolors.OKBLUE}{path_to_custom_metadata}{bcolors.ENDC}'
        )
        with open(path_to_custom_metadata, 'r') as f:
            custom_metadata = json.load(f)

    # set the number of workers if unset
    if cfg['loader']['num_workers'] < 0:
        # set the number of workers to the number of CPUs available,
        # but minimum of 8
        num_workers = max(8, cpu_count())
        num_workers = min(32, num_workers)
        cfg['loader']['num_workers'] = num_workers

    size = cfg['resize']
    if not isinstance(size, int):
        size = tuple(size)
    transform = None
    if isinstance(size, tuple) or size > 0:
        transform = torchvision.transforms.Resize(size)

    if input_dir:
        if not cfg.append and len(api_workflow_client.get_all_tags()) > 0:
            print_as_warning(
                'The dataset you specified already has samples. '
                'If you want to add additional samples, you need to specify '
                'append=True as CLI argument.'
            )
            return

        mode = cfg['upload']
        dataset = LightlyDataset(input_dir=input_dir, transform=transform)
        api_workflow_client.upload_dataset(
            input=dataset,
            mode=mode,
            max_workers=cfg['loader']['num_workers'],
            custom_metadata=custom_metadata,
        )
        print('Finished the upload of the dataset.')

    if path_to_embeddings:
        name = cfg['embedding_name']
        if not cfg.append:
            try:
                embedding = api_workflow_client.get_embedding_by_name(name=name, ignore_suffix=True)
                print_as_warning(
                    'The dataset you specified already has an embedding. '
                    'If you want to add additional samples, you need to specify '
                    'append=True as CLI argument.'
                )
                return
            except EmbeddingDoesNotExistError:
                pass
        api_workflow_client.upload_embeddings(
            path_to_embeddings_csv=path_to_embeddings, name=name
        )
        print('Finished upload of embeddings.')

    if custom_metadata is not None and not input_dir:
        # upload custom metadata separately
        api_workflow_client.upload_custom_metadata(
            custom_metadata,
            verbose=True,
            max_workers=cfg['loader']['num_workers'],
        )

    if new_dataset_name:
        print(f'The dataset_id of the newly created dataset is '
              f'{bcolors.OKBLUE}{api_workflow_client.dataset_id}{bcolors.ENDC}')

    os.environ[
        cfg['environment_variable_names']['lightly_last_dataset_id']
    ] = api_workflow_client.dataset_id

    return SUCCESS_RETURN_VALUE


@hydra.main(**fix_hydra_arguments(config_path = 'config', config_name = 'config'))
def upload_cli(cfg):
    """Upload images/embeddings from the command-line to the Lightly platform.

    Args:
        cfg:
            The default configs are loaded from the config file.
            To overwrite them please see the section on the config file
            (.config.config.yaml).

    Command-Line Args:
        input_dir:
            Path to the input directory where images are stored.
        embeddings:
            Path to the csv file storing the embeddings generated by
            lightly.
        token:
            User access token to the Lightly platform. If needs to be
            specified to upload the images and embeddings to the platform.
        dataset_id:
            Identifier of the dataset on the Lightly platform.
            Either the dataset_id or the new_dataset_name need to be
            specified.
        new_dataset_name:
            The name of the new dataset to create on the Lightly platform.
            Either the dataset_id or the new_dataset_name need to be
            specified.
        upload:
            String to determine whether to upload the full images,
            thumbnails only, or metadata only.

            Must be one of ['full', 'thumbnails', 'metadata']
        embedding_name:
            Assign the embedding a name in order to identify it on the
            Lightly platform.
        resize:
            Desired size of the uploaded images. If negative, default size is
            used. If size is a sequence like (h, w), output size will be matched
            to this. If size is an int, smaller edge of the image will be
            matched to this number. i.e, if height > width, then image will be
            rescaled to (size * height / width, size).
        custom_metadata:
            Path to a .json file containing custom metadata. The file must be in
            the COCO annotations (although annotations can be empty) format and
            contain an additional field `metadata` storing a list of metadata
            entries. The metadata entries are matched with the images via
            `image_id`.

    Examples:
        >>> # create a new dataset on the Lightly platform and upload full images to it
        >>> lightly-upload input_dir=data/ token='123' new_dataset_name='new_dataset_name_xyz'
        >>>
        >>> # upload full images to the Lightly platform to an existing dataset
        >>> lightly-upload input_dir=data/ token='123' dataset_id='XYZ'
        >>>
        >>> # create a new dataset on the Lightly platform and upload thumbnails to it
        >>> lightly-upload input_dir=data/ token='123' new_dataset_name='new_dataset_name_xyz' upload='thumbnails'
        >>>
        >>> # upload metadata to the Lightly platform
        >>> lightly-upload input_dir=data/ token='123' dataset_id='XYZ' upload='metadata'
        >>>
        >>> # upload embeddings to the Lightly platform (must have uploaded images beforehand)
        >>> lightly-upload embeddings=embeddings.csv token='123' dataset_id='XYZ'
        >>>
        >>> # upload both, images and embeddings in a single command
        >>> lightly-upload input_dir=data/ embeddings=embeddings.csv upload='full' \\
        >>>     token='123' dataset_id='XYZ'
        >>>
        >>> # create a new dataset on the Lightly platform and upload both, images and embeddings
        >>> lightly-upload input_dir=data/ embeddings=embeddings.csv upload='full' \\
        >>>     token='123' new_dataset_name='new_dataset_name_xyz'
        >>>
        >>> # upload a dataset with custom metadata
        >>> lightly-upload input_dir=data/ token='123' dataset_id='XYZ' custom_metadata=custom_metadata.json
        >>>
        >>> # upload custom metadata to an existing dataset
        >>> lightly-upload token='123' dataset_id='XYZ' custom_metadata=custom_metadata.json

    """
    return _upload_cli(cfg)


def entry():
    upload_cli()



""" Collection of Active Learning Scorers """

# Copyright (c) 2020. Lightly AG and its affiliates.
# All Rights Reserved

from lightly.active_learning.scorers.scorer import Scorer
from lightly.active_learning.scorers.classification import ScorerClassification
from lightly.active_learning.scorers.detection import ScorerObjectDetection
from lightly.active_learning.scorers.semantic_segmentation import ScorerSemanticSegmentation



import torch
import numpy as np


def cosine_schedule(
    step: int, max_steps: int, start_value: float, end_value: float
) -> float:

    """
    Use cosine decay to gradually modify start_value to reach target end_value during iterations.

    Args:
        step:
            Current step number.
        max_steps:
            Total number of steps.
        start_value:
            Starting value.
        end_value:
            Target value.

    Returns:
        Cosine decay value.

    """
    if step < 0:
        raise ValueError("Current step number can't be negative")
    if max_steps < 1:
        raise ValueError("Total step number must be >= 1")
    if step >= max_steps:
        raise ValueError(
            f"The current step must be smaller than max_steps but found step equal to {step} and max_steps equal to {max_steps}."
        )

    if max_steps == 1:
        decay = end_value
    else:
        decay = (
            end_value
            - (end_value - start_value)
            * (np.cos(np.pi * step / (max_steps - 1)) + 1)
            / 2
        )
    return decay


class CosineWarmupScheduler(torch.optim.lr_scheduler.LambdaLR):
    """
    Cosine warmup scheduler for learning rate.

    Args:
        optimizer:
            Optimizer object to schedule the learning rate.
        warmup_epochs:
            Number of warmup epochs.
        max_epochs:
            Total number of training epochs.
        last_epoch:
            The index of last epoch. Default: -1
        verbose:
            If True, prints a message to stdout for each update. Default: False.
    """

    def __init__(
        self,
        optimizer: torch.optim.Optimizer,
        warmup_epochs: int,
        max_epochs: int,
        last_epoch: int = -1,
        verbose: bool = False,
    ) -> None:
        self.warmup_epochs = warmup_epochs
        self.max_epochs = max_epochs
        super().__init__(
            optimizer=optimizer,
            lr_lambda=self.scale_lr,
            last_epoch=last_epoch,
            verbose=verbose,
        )

    def scale_lr(self, epoch: int) -> float:
        """
        Scale learning rate according to the current epoch number.

        Args:
            epoch:
                Current epoch number.

        Returns:
            Scaled learning rate.

        """
        if epoch < self.warmup_epochs:
            return epoch / self.warmup_epochs
        else:
            return cosine_schedule(
                epoch - self.warmup_epochs, self.max_epochs - self.warmup_epochs, 0, 1
            )
